import numpy as np
import pandas as pd
import soundfile as sf
import os
from pathlib import Path



def find_overlap(row, df2, filename_col='filename', start_col='start', end_col='end'):
    """
    Check if a given row overlaps temporally (time) with any row in another DataFrame (df2).

    Args:
        row: pd.Series
            A single row (pandas Series) from a DataFrame that contains the 'start' and 'end' columns,
            as well as a 'filename' column that specifies the audio file to which the row belongs.   
        df2: pd.DataFrame
            A DataFrame containing rows with 'start' and 'end' columns, as well as a 'filename' column.
            This DataFrame is checked for overlap against the given row.
        filename_col: str
            The name of the column in both dataframes that contains the filename. 
            Defaults to 'filename'.
        start_col: str
            The name of the column in both dataframes that contains the start time. 
            Defaults to 'start'.
        end_col: str
            The name of the column in both dataframes that contains the end time. 
            Defaults to 'end'.

    Returns: bool
        Returns True if the given row overlaps with any row in df2 for the same filename,
        otherwise returns False.

    Example:

    >>> import pandas as pd
    >>> df1 = pd.DataFrame({
    ... 'filename': ['file1', 'file2'],
    ... 'start': [1, 3],
    ... 'end': [2, 4]
    ... })
    >>> df2 = pd.DataFrame({
    ... 'filename': ['file1', 'file1', 'file2'],
    ... 'start': [1.5, 2.5, 3.5],
    ... 'end': [2.5, 3.5, 4.5]
    ... })
    >>> row = df1.iloc[0]
    >>> find_overlap(row, df2)
    True
    
    >>> row = df1.iloc[1]
    >>> find_overlap(row, df2)
    True
    """

    interval1 = pd.Interval(row[start_col], row[end_col])
    
    matching_df2 = df2[df2[filename_col] == row[filename_col]]
    
    for _, row2 in matching_df2.iterrows():
        interval2 = pd.Interval(row2[start_col], row2[end_col])
        if interval1.overlaps(interval2):
            return True
                
    return False

def get_continuous_results(evaluation, reference, threshold=0.5):
    """
    Evaluate the True Positives (TP), False Positives (FP), and False Negatives (FN) for each unique class label
    in the ground truth DataFrame based on the evaluation DataFrame.

    This function is suitable for long, continuous segments of data and is designed to compare two sets of 
    time-stamped annotations, which could either be predictions and ground truths, or two different sets of annotations.
    
    Args:
        evaluation: pd.DataFrame
            DataFrame containing the evaluation results, which could be detection scores or another set of annotations.
            The DataFrame should have columns like 'label', 'start', 'end', etc.
        reference: pd.DataFrame
            DataFrame containing the ground truth labels, with similar structure to `predicted`.
        threshold: float
            The threshold for classifying a detection as positive or negative.
    
    Returns: dict
        A dictionary where keys are the unique class labels, and the values are dictionaries containing the counts of TP, FP, FN.
    """
    # Filter detections based on the threshold
    predictions_threshold = evaluation[evaluation['score'] >= threshold]
    unique_classes = reference['label'].unique().tolist()
    evaluation_results = {}
   
    for cls in unique_classes:
        tp = fp = fn = 0
        pred_filtered = predictions_threshold[predictions_threshold['label'] == cls]
        gt_filtered = reference[reference['label'] == cls]
        # We are now going to calculate the TP and FN. Note that this calculation covers instances where we have one prediction that overlaps with
        # multiple ground truths. In this case, the TP will be incremented for each ground truth
        # We are also considering Multiple preditcions for one ground truth, in this case, only one of the predictions will increment the TP.
        # Another assumption made here is that even if only 1% of the prediction falls within the ground truth it will count as a TP

        # Loop through each ground truth entry
        for _, gt_row in gt_filtered.iterrows():
            # If an overlap is found, increment TP, otherwise increment FN
            if find_overlap(gt_row, pred_filtered):
                tp += 1
            else:
                fn += 1
        
        # Calculate FP:
        # Loop through each predicted filtered entry
        for _, predicted_row in pred_filtered.iterrows():
            # If no overlap is found, increment FP
            if not find_overlap(predicted_row, gt_filtered):
                fp += 1

        evaluation_results[cls] = {
            'TP': tp,
            'FP': fp,
            'FN': fn
        }

    return evaluation_results

def load_file_list(file_path):
    """
    Load a list of file paths from a text file.

    Args:
        file_path: str
            Path to the text file containing file paths.

    Returns:
        list of str: List containing file paths.
    """
    file_list = []
    with open(file_path, 'r') as file:
        for line in file:
            # Strip to remove leading/trailing whitespaces and newline characters
            clean_line = line.strip()
            if clean_line:  # Check if line is not empty
                file_list.append(clean_line)
    return file_list

def create_background_annotations(annotations, file_list, label=0, root_audio_folder=None):
    """
    Create background annotations for audio files given a set of existing annotations.
    
    This function scans through each audio file in the provided list and its annotations. 
    For periods where no annotations exist, it generates "background" annotations.
    
    Args:
        annotations: pd.DataFrame
            DataFrame containing the existing annotations. The DataFrame should have columns ['filename', 'start', 'end', 'label'].
        file_list: list of str
            List of paths to the audio files.
        root_audio_folder: str
            Path to the a root folder for the audio files. This folder may contain subfolders that are specified in each annotation file path.
        label: int
            The label to assign to the background annotations. Default is 0.
    
    Returns:
        pd.DataFrame: A new DataFrame containing both the original and background annotations.
    """
    background_annotations = []
    
    for file_path in file_list:
        full_file_path = os.path.join(root_audio_folder, file_path) if root_audio_folder else file_path
        audio_duration = sf.info(full_file_path).duration
        
        # Extract all annotations for the current file
        file_annotations = annotations[annotations['filename'] == file_path].sort_values(by='start')
        
        # If there are no annotations for this file, create a single background annotation for the entire duration
        if file_annotations.empty:
            background_annotations.append({
                'filename': file_path,
                'start': 0,
                'end': audio_duration,
                'label': label
            })
        else:
            # Initial background start time
            last_end_time = 0
            
            for _, row in file_annotations.iterrows():
                # If there is a gap between the last annotation's end and the current annotation's start
                if last_end_time < row['start']:
                    background_annotations.append({
                        'filename': file_path,
                        'start': last_end_time,
                        'end': row['start'],
                        'label': label
                    })
                last_end_time = row['end']
            
            # Check for background after the last annotation
            if last_end_time < audio_duration:
                background_annotations.append({
                    'filename': file_path,
                    'start': last_end_time,
                    'end': audio_duration,
                    'label': label
                })
    
    # Convert to DataFrame
    background_df = pd.DataFrame(background_annotations)
    
    # Concatenate with original annotations and return
    return pd.concat([annotations, background_df], ignore_index=True)


def get_clips_results(evaluation, reference, threshold=0.5):
    """
    Calculate True Positives (TP), False Positives (FP), False Negatives (FN), and True Negatives (TN)
    for each class at a given threshold using a One-vs-All (OvA) approach.

    In the OvA approach, each class is considered as the 'positive' class, while all other classes are
    grouped into the 'negative' class. Metrics are then calculated for this binary classification problem
    and are done so for each class.

    This implementation also assumes that filenames are unique.

    Args:
        evaluation: Pandas DataFrame
            DataFrame containing evaluation results, which could be detection scores or a set of annotations. 
            The DataFrame should have columns: ['filename', 'label', 'score'].
        reference: Pandas DataFrame
            DataFrame containing ground-truth reference with columns: ['filename', 'label']
        threshold: float
            The threshold for classifying a detection as positive or negative.

    Returns:
        dict: A dictionary containing counts of TPs, FPs, FNs, and TNs for each class.
            The keys of the dictionary are the unique class labels, and the values are dictionaries containing
            the counts of TPs, FPs, FNs, and TNs for that class.

    """
    # Initialize counts
    counts = {}

    # Filter detections based on the threshold
    filtered_detections = evaluation[evaluation['score'] >= threshold]

    # Iterate through unique labels
    for label in reference['label'].unique():
        # Initialize TP and FP for this label
        TP = 0
        FP = 0
        FN = 0
        TN = 0

        # Find instances where the ground truth is this label
        gt_label = reference[reference['label'] == label]

        # Find instances where the detection is this label and above the threshold
        det_label = filtered_detections[filtered_detections['label'] == label]

        # Calculate TP and FP
        # Create a mask of all detections that have a corresponding entry in the gt_label
        TP = len(det_label[det_label['filename'].isin(gt_label['filename'])])
        FP = len(det_label) - TP

        # Calculate FN
        # this will create a mask of all gt that do not have a corresponding entry in det_label
        FN = len(gt_label[~gt_label['filename'].isin(det_label['filename'])])

        # selects all the ground truth instances that do not belong to the current label of interest. These are the potential True Negatives. 
        gt_not_label = reference[reference['label'] != label]
        # filters out the filenames that are present in the detection list for the current label. The remaining instances are correctly classified as not belonging to the current class
        TN = len(gt_not_label[~gt_not_label['filename'].isin(det_label['filename'])])

        # Store results
        counts[label] = {'TP': TP, 'FP': FP, 'FN': FN, 'TN': TN}

    return counts

def calculate_metrics(TP, FP, FN, TN=None, total_time_units=None):
    """
    Calculate classification metrics like precision, recall, F1-score, 
    and optionally accuracy and False Positive Rate (FPR) per time unit.

    Args:
        TP: int 
            Number of True Positives
        FP: int
            Number of False Positives
        FN: int
            Number of False Negatives
        TN: int
            Number of True Negatives. If None, will not calculate metrics that need TN.
        total_time_units: float
            The total duration in an arbitrary unit of time (e.g., hours, minutes). 
            If provided, will calculate FPR per time unit.

    Returns:
        dict: A dictionary containing the calculated metrics.
    """
    # Initialize a dictionary to store the metrics
    metrics = {}

    # Calculate precision
    if TP + FP == 0:
        metrics['Precision'] = 0
    else:
        metrics['Precision'] = TP / (TP + FP)

    # Calculate recall
    if TP + FN == 0:
        metrics['Recall'] = 0
    else:
        metrics['Recall'] = TP / (TP + FN)

    # Calculate F1-score
    if metrics['Precision'] + metrics['Recall'] == 0:
        metrics['F1-Score'] = 0
    else:
        metrics['F1-Score'] = 2 * metrics['Precision'] * metrics['Recall'] / (metrics['Precision'] + metrics['Recall'])

    # Calculate accuracy
    if TN is not None:
        if TP + FP + TN + FN == 0:
            metrics['Accuracy'] = 0
        else:
            metrics['Accuracy'] = (TP + TN) / (TP + FP + TN + FN)

    # Calculate FPR per hour if total_time_units is provided
    if total_time_units is not None:
        if total_time_units == 0:
            metrics['FPR_per_time_unit'] = 0
        else:
            metrics['FPR_per_time_unit'] = FP / total_time_units


    return metrics

def evaluate(evaluation, reference, type='continuous', threshold_min=0, threshold_max=1, threshold_inc=0.05, total_time_units=None, output_folder=None, add_background_reference=None):
    """
    Evaluate the performance of sound detections for both shorter clips and longer continuous files.
    This function iterates over a range of detection score thresholds to evaluate performance metrics 
    like True Positive, True Negative, False Positive, and False Negative counts for each class.
    Depending on the 'type' specified, additional metrics may be calculated. 
    It also calculates macro and micro averages for metrics like Precision, Recall, and F1-Score.

    Args:
        evaluation: str
            Path to the CSV file containing evaluation results, which could be detection scores or a set of annotations.
        reference: str 
            Path to the CSV file containing ground-truth reference.
        type: str
            Type of audio data being evaluated. Accepts 'continuous' for long continuous files or 
            'clips' for shorter audio clips. Affects which metrics are calculated. Defaults to 'continuous'.
        threshold_min: float
            Minimum threshold for detection. Defaults to 0.
        threshold_max: float
            Maximum threshold for detection. Defaults to 1.
        threshold_inc: float
            Threshold increment for each step. Defaults to 0.05.
        total_time_units: float or None
            The total duration in arbitrary time units over which the detections were made (e.g., hours, minutes).
            This is necessary to calculate the False Positive Rate per unit time. Defaults to None.
        output_folder: str or None
            The folder where to save the output CSV files. If None, the current directory is used.
            Defaults to None.
        add_background_reference: tuple or None
            Tuple containing the root audio folder with the files you want to add background reference to, a path to a file list 
            containing the path to each individual file from teh root folder and the label to create background reference,
            or None if no background reference should be created. Only relevant when 'type' is 'continuous'.
            Defaults to None.

    Returns:
        tuple: Two Pandas DataFrames.
            - The first DataFrame contains detection results for each threshold and class.
            - The second DataFrame contains classification metrics for each threshold and class, also including macro and micro averages.
    """
    # Validate 'type' argument
    if type not in ['continuous', 'clips']:
        raise ValueError("Invalid value for 'type'. It should be either 'continuous' or 'clips'.")

    # Initialize dictionaries to store results and metrics for different thresholds
    all_threshold_results = {}
    all_threshold_metrics = {}

    # Read the detection and annotation files
    evaluation = pd.read_csv(evaluation)
    reference = pd.read_csv(reference)

    # Determine the output folder
    if output_folder is None:
        output_folder = Path('.').resolve()
    else:
        output_folder = Path(output_folder).resolve()

    output_folder.mkdir(parents=True, exist_ok=True)

    if add_background_reference is not None and type == 'continuous':
        root_audio_folder = None
        if len(add_background_reference) == 3:
            root_audio_folder = add_background_reference[2]
        path_to_audio_list = add_background_reference[0]
        file_list = load_file_list(path_to_audio_list)
        label = int(add_background_reference[1])
        reference = create_background_annotations(reference, file_list, root_audio_folder=root_audio_folder, label=label)

    # List of metrics to calculate
    metrics_to_calculate = ['Precision', 'Recall', 'F1-Score'] # These metrics will always be calculated no matter the type
    
    # Include 'Accuracy' if the type is 'clips'
    if type == 'clips':
        metrics_to_calculate.append('Accuracy')

    if total_time_units is not None:
        metrics_to_calculate.append('FPR_per_time_unit')


    if threshold_min == threshold_max:
        thresholds = [threshold_min]
    else:
        thresholds = list(np.arange(threshold_min, threshold_max, threshold_inc))
        if thresholds[-1] != threshold_max: # Manually including the last threshold
            thresholds.append(threshold_max)
    
    for threshold in thresholds:
        threshold = round(threshold, 5) # rounding the threshold to avoid floating-point issues
        
        # Get results
        if type == 'clips':
            result = get_clips_results(evaluation, reference, threshold=threshold)
        else:
            result = get_continuous_results(evaluation, reference, threshold=threshold)
        metrics_for_this_threshold = {}
    
        total_TP = 0
        total_FP = 0
        total_FN = 0
        total_TN = 0
        
        # Loop through each unique class label to calculate metrics
        for label, counts in result.items():
            TP = counts['TP']
            FP = counts['FP']
            FN = counts['FN']
            
            total_TP += TP
            total_FP += FP
            total_FN += FN

            if type == 'clips':
                TN = counts['TN']
                total_TN += TN
            else:
                TN = None
                total_TN = None
            
            metrics = calculate_metrics(TP, FP, FN, TN, total_time_units=total_time_units)
            metrics_for_this_threshold[label] = metrics
        
        
        # Calculate and store macro-average metrics
        macro_avg_metrics = {metric: np.mean([x[metric] for x in metrics_for_this_threshold.values()])
                            for metric in metrics_to_calculate}
        
        metrics_for_this_threshold['macro_avg'] = macro_avg_metrics  # Store macro averages as a 'label' in the result dictionary
        
        # Calculate and store micro-average metrics
        micro_avg_metrics = calculate_metrics(total_TP, total_FP, total_FN, total_TN, total_time_units=total_time_units)
        metrics_for_this_threshold['micro_avg'] = micro_avg_metrics  # Store micro averages as a 'label' in the result dictionary

        all_threshold_metrics[threshold] = metrics_for_this_threshold
        # Store the results and metrics for this threshold
        all_threshold_results[threshold] = result

    # Convert these nested dictionaries to a Pandas DataFrame
    all_results_df = pd.DataFrame.from_dict({(i, j): all_threshold_results[i][j] 
                                             for i in all_threshold_results.keys() 
                                             for j in all_threshold_results[i].keys()}, 
                                             orient='index')

    all_metrics_df = pd.DataFrame.from_dict({(i, j): all_threshold_metrics[i][j] 
                                             for i in all_threshold_metrics.keys() 
                                             for j in all_threshold_metrics[i].keys()}, 
                                             orient='index')
    
    # Name the multi-level index columns
    all_results_df.index.names = ['threshold', 'class']
    all_metrics_df.index.names = ['threshold', 'class']

    # Save these DataFrames to CSV
    all_results_df.to_csv((output_folder / 'results.csv'), index=True)
    all_metrics_df.to_csv((output_folder / 'metrics.csv'), index=True)

    print("All results and metrics saved to {}".format(output_folder))
    
    return all_results_df, all_metrics_df

def main():
    import argparse
    import warnings
    import ast

    def boolean_string(s):
            if s not in {'False', 'True'}:
                raise ValueError('Not a valid boolean string')
            return s == 'True'

    def tryeval(val):
        # Literal eval does cast type safely. However, doesnt work for str, therefore the try except.
        try:
            val = ast.literal_eval(val)
        except ValueError:
            pass
        return val
        
    class ParseKwargs(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            setattr(namespace, self.dest, dict())
            for value in values:
                key, value = value.split('=')
                getattr(namespace, self.dest)[key] = tryeval(value)

    parser = argparse.ArgumentParser(description="Evaluate the performance of sound detection.")
    parser.add_argument('evaluation', type=str, help='Path to the .csv file containing evaluation results, which may include detection scores or annotations.')
    parser.add_argument('reference', type=str, help='Path to the .csv file containing ground truth reference.')
    parser.add_argument('--type', choices=['clips', 'continuous'], default='continuous', help='Type of evaluation: "clips" for short clips or "continuous" for long continuous files.')
    parser.add_argument('--threshold_min', default=0, type=float, help='Minimum threshold for detection.')
    parser.add_argument('--threshold_max', default=1, type=float, help='Maximum threshold for detection.')
    parser.add_argument('--threshold_inc', default=0.05, type=float, help='Threshold increment for each step.')
    parser.add_argument('--total_time_units', default=None, type=float, help='The total duration in arbitrary time units over which the detections were made (e.g., hours, minutes).' \
            ' This is necessary to calculate the False Positive Rate per unit time. Defaults to None.')
    parser.add_argument('--output_folder', default=None, type=str, help='Location to output the performance results. For instance: metrics/')
    parser.add_argument('--add_background_reference', nargs='*', default=None, help=(
            "Create background reference annotations for audio files based on existing annotations. "
            "This option is only relevant for the 'continuous' data type. Up to three parameters can be passed in the following order: "
            "[path_to_file_list, label, root_audio_folder]. "
            "'path_to_file_list' is a required parameter specifying the path to a text file that contains a list of audio file paths, one per line. "
            "'label' is an optional parameter specifying the label to assign to background annotations (default is 0). "
            "'root_audio_folder' is an optional path to the root directory containing the audio files, which will be prepended to each path in the file list. "
            "If omitted, only the paths from the file list are used. "
            "If only one parameter is provided, it is assumed to be 'path_to_file_list'. If two are provided, they are assumed to be 'path_to_file_list' and 'label'."
            ))

    args = parser.parse_args()

    # Manually process the second part of --add_background_annot
    if hasattr(args, 'add_background_reference') and args.add_background_reference:
        args.add_background_reference[2] = tryeval(args.add_background_reference[2])

    args = parser.parse_args()
    evaluate(**vars(args))

if __name__ == "__main__":
    main()


# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Unit tests for the data_feeding module within the ketos library
"""

import os
import pytest
import warnings
import numpy as np
from tables import open_file
from ketos.data_handling.database_interface import open_table
from ketos.data_handling.data_feeding import BatchGenerator, JointBatchGen, MultiModalBatchGen
from ketos.neural_networks.resnet import ResNetInterface


current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(os.path.dirname(current_dir),"assets")
path_to_tmp = os.path.join(path_to_assets,'tmp')

def test_one_batch():
    """ Test if one batch has the expected shape and contents
    """
    h5 = open_file(os.path.join(path_to_assets, "mini_narw.h5"), 'r') # create the database handle  
    train_data = open_table(h5, "/train/data")

    five_specs = train_data[:5]['data']
    five_labels = train_data[:5]['label']
    
    five_labels = [np.array(l) for l in five_labels]

    train_generator = BatchGenerator(data_table=train_data,batch_size=5, return_batch_ids=True) #create a batch generator 
    ids, X, Y = next(train_generator)

    np.testing.assert_array_equal(ids,[0,1,2,3,4])
    assert X.shape == (5, 94, 129)
    np.testing.assert_array_equal(X, five_specs)
    assert Y.shape == (5,)
    np.testing.assert_array_equal(Y, five_labels)

    h5.close()

def test_multiple_data_fields():
    """ Test if one batch has the expected shape and contents when loading multiple data fields 
        from the same table
    """
    h5 = open_file(os.path.join(path_to_assets, "mini_narw_mult.h5"), 'r') # create the database handle  
    train_data = open_table(h5, "/train/data")

    five_specs = train_data[:5]['spec']
    five_gammas = train_data[:5]['gamma']
    five_labels = train_data[:5]['label']

    five_labels = [np.array(l) for l in five_labels]

    train_generator = BatchGenerator(data_table=train_data, batch_size=5, x_field=['spec','gamma'], return_batch_ids=True) #create a batch generator 
    ids, X, Y = next(train_generator)
    
    np.testing.assert_array_equal(ids,[0,1,2,3,4])
    assert len(X) == 5
    assert len(X[0]) == 2
    assert X[0]['spec'].shape == (94, 129)
    assert X[0]['gamma'].shape == (3000, 20)
    specs = [x[0] for x in X]
    gammas = [x[1] for x in X]
    np.testing.assert_array_equal(specs, five_specs)
    np.testing.assert_array_equal(gammas, five_gammas)
    assert Y.shape == (5,)
    np.testing.assert_array_equal(Y, five_labels)

    h5.close()

def test_output_for_strong_annotations():
    """ Test if batch generator returns multiple labels for strongly annotated instances
    """
    h5 = open_file(os.path.join(path_to_assets, "11x_same_spec.h5"), 'r') # create the database handle  
    data = open_table(h5, "/group_1/table_data")
    annot = open_table(h5, "/group_1/table_annot")

    expected_y = np.array([[annot[0]['label'],annot[1]['label']],
                            [annot[2]['label'],annot[3]['label']],
                            [annot[4]['label'],annot[5]['label']],
                            [annot[6]['label'],annot[7]['label']],
                            [annot[8]['label'],annot[9]['label']]])
                            
    train_generator = BatchGenerator(batch_size=5, data_table=data, annot_in_data_table=False, 
        annot_table=annot, y_field=['label'], shuffle=False, refresh_on_epoch_end=False)
    
    _, Y = next(train_generator)
    np.testing.assert_array_equal(Y, expected_y)

    h5.close()


def test_output_for_strong_annotations_not_all_samples_have_annotations():
    """ Test if batch generator returns multiple labels for strongly annotated instances, when 
        some samples have no annotations
    """
    h5 = open_file(os.path.join(path_to_assets, "11x_wf_annot_tbl.h5"), 'r') # create the database handle  
    data = open_table(h5, "/audio")
    annot = open_table(h5, "/audio_annot")

    # first, using only one y-field
    expected_y = np.array([[0, 1], [2], [], [0, 1]])                            
    train_generator = BatchGenerator(batch_size=4, data_table=data, annot_in_data_table=False, 
        annot_table=annot, y_field=['label'], shuffle=False, refresh_on_epoch_end=False)
    _, Y = next(train_generator)
    np.testing.assert_array_equal(Y, expected_y)

    # first, now let's try with two y-fields
    dtype = {'names':['label','end'], 'formats':['u1','<f8'], 'offsets':[20,4], 'itemsize':29}
    expected_y = [np.array([(0, 2.2), (1, 4.2)], dtype=dtype), 
                  np.array([(2, 6.2)], dtype=dtype), 
                  np.array([], dtype=dtype), 
                  np.array([(0, 2.2), (1, 4.2)], dtype=dtype)]                            
    train_generator = BatchGenerator(batch_size=4, data_table=data, annot_in_data_table=False, 
        annot_table=annot, y_field=['label','end'], shuffle=False, refresh_on_epoch_end=False)
    _, Y = next(train_generator)
    for _Y,_y in zip(Y, expected_y):
        np.testing.assert_array_equal(_Y, _y)

    h5.close()


def test_batch_sequence_same_as_db():
    """ Test if batches are generated with instances in the same order as they appear in the database
    """
    h5 = open_file(os.path.join(path_to_assets, "mini_narw.h5"), 'r') # create the database handle  
    train_data = open_table(h5, "/train/data")

    ids_in_db = train_data[:]['id']
    train_generator = BatchGenerator(data_table=train_data, batch_size=3, return_batch_ids=True) #create a batch generator 

    for i in range(3):
        ids, X, _ = next(train_generator)
        np.testing.assert_array_equal(X, train_data[ids_in_db[i*3: i*3+3]]['data'])
        np.testing.assert_array_equal(ids,list(range(i*3, i*3+3)))
    
    h5.close()


def test_last_batch():
    """ Test if last batch has the expected number of instances
    """
    h5 = open_file(os.path.join(path_to_assets, "mini_narw.h5"), 'r') # create the database handle  
    train_data = open_table(h5, "/train/data")

    ids_in_db = train_data[:]['id']
    train_generator = BatchGenerator(data_table=train_data, batch_size=6, return_batch_ids=True) #create a batch generator 
    #First batch
    ids, X, _ = next(train_generator)
    np.testing.assert_array_equal(ids,[0,1,2,3,4,5])
    assert X.shape == (6, 94, 129)
    #Second batch
    ids, X, _ = next(train_generator)
    np.testing.assert_array_equal(ids,[6,7,8,9,10,11])
    assert X.shape == (6, 94, 129)

    #Third batch; Last batch ( will have the remaining instances)
    ids, X, _ = next(train_generator)
    np.testing.assert_array_equal(ids,[12, 13, 14, 15, 16, 17, 18, 19])
    assert X.shape == (8, 94, 129)
    
    h5.close()

def test_use_only_subset_of_data():
    """ Test that only the indices specified are used
    """
    h5 = open_file(os.path.join(path_to_assets, "mini_narw.h5"), 'r') # create the database handle  
    train_data = open_table(h5, "/train/data")
    
    train_generator = BatchGenerator(data_table=train_data, batch_size=4, select_indices=[1,3,5,7,9,11,13,14], return_batch_ids=True) #create a batch generator 
    #First batch
    ids, X, _ = next(train_generator)
    np.testing.assert_array_equal(ids,[1,3,5,7])
    #Second batch
    ids, X, _ = next(train_generator)
    np.testing.assert_array_equal(ids,[9,11,13,14])

    h5.close()

def test_multiple_epochs():
    """ Test if batches are as expected after the first epoch
    """
    h5 = open_file(os.path.join(path_to_assets, "mini_narw.h5"), 'r') # create the database handle  
    train_data = open_table(h5, "/train/data")

    ids_in_db = train_data[:]['id']
    train_generator = BatchGenerator(data_table=train_data, batch_size=6, return_batch_ids=True) #create a batch generator 
    #Epoch 0, batch 0
    ids, X, _ = next(train_generator)
    np.testing.assert_array_equal(ids,[0,1,2,3,4,5])
    assert X.shape == (6, 94, 129)
    #Epoch 0, batch 1
    ids, X, _ = next(train_generator)
    np.testing.assert_array_equal(ids,[6,7,8,9,10,11])
    assert X.shape == (6, 94, 129)

    ##Epoch 0, batch 2 Last batch ( will have the remaining instances)
    ids, X, _ = next(train_generator)
    np.testing.assert_array_equal(ids,[12, 13, 14, 15, 16, 17, 18, 19])
    assert X.shape == (8, 94, 129)
    
    #Epoch 1, batch 0
    ids, X, _ = next(train_generator)
    np.testing.assert_array_equal(ids,[0,1,2,3,4,5])
    assert X.shape == (6, 94, 129)

    h5.close()

def test_load_from_memory():
    """ Test if batch generator can work with data loaded from memory
    """
    x = np.ones(shape=(15,32,16))
    y = np.zeros(shape=(15))

    generator = BatchGenerator(x=x, y=y, batch_size=6, return_batch_ids=True) #create a batch generator 

    #Epoch 0, batch 0
    ids, X, _ = next(generator)
    assert ids == [0,1,2,3,4,5]
    assert X.shape == (6, 32, 16)
    #Epoch 0, batch 1
    ids, X, _ = next(generator)
    assert ids == [6,7,8,9,10,11,12,13,14]
    assert X.shape == (9, 32, 16)
    
    
    #Epoch 1, batch 0
    ids, X, _ = next(generator)
    assert ids == [0,1,2,3,4,5]
    assert X.shape == (6, 32, 16)

def test_shuffle():
    """Test shuffle argument.
        Instances should be shuffled before divided into batches, but the order should be consistent across epochs if
        'refresh_on_epoch_end' is False.
    """
    h5 = open_file(os.path.join(path_to_assets, "mini_narw.h5"), 'r') # create the database handle  
    train_data = open_table(h5, "/train/data")

    np.random.seed(100)

    ids_in_db = train_data[:]['id']
    train_generator = BatchGenerator(data_table=train_data, batch_size=6, return_batch_ids=True, shuffle=True) #create a batch generator 

    
    for epoch in range(5):
        #batch 0
        ids, X, _ = next(train_generator)
        np.testing.assert_array_equal(ids,[17, 19, 11, 18, 13,  6])
        assert X.shape == (6,94,129)
        #batch 1
        ids, X, _ = next(train_generator)
        np.testing.assert_array_equal(ids, [16, 1, 9, 14, 12, 5])
        assert X.shape == (6, 94, 129)
        #batch 2
        ids, X, _ = next(train_generator)
        np.testing.assert_array_equal(ids,[2, 4, 10, 0, 15, 7, 3, 8])
        assert X.shape == (8, 94, 129)

       
    h5.close()


def test_refresh_on_epoch_end():
    """ Test if batches are generated with randomly selected instances for each epoch
    """
    h5 = open_file(os.path.join(path_to_assets, "mini_narw.h5"), 'r') # create the database handle  
    train_data = open_table(h5, "/train/data")

    np.random.seed(100)

    ids_in_db = train_data[:]['id']
    train_generator = BatchGenerator(data_table=train_data, batch_size=6, return_batch_ids=True, shuffle=True, refresh_on_epoch_end=True) #create a batch generator 

    expected_ids = {'epoch_1': ([17, 19, 11, 18, 13,  6], [16,  1,  9, 14, 12,  5], [2,  4, 10,  0, 15, 7,  3,  8]),    
                     'epoch_2':  ([18, 19, 17, 0, 8, 6], [14, 7, 11, 10, 15, 3], [5, 13, 1, 4, 12, 2, 9, 16]),
                     'epoch_3': ([3, 4, 12, 17, 10, 1], [19, 5, 11, 8, 0, 18], [6, 13, 7, 15, 16, 14, 2, 9])}
                     
    for epoch in ['epoch_1', 'epoch_2', 'epoch_3']:
        #batch 0
        ids, X, _ = next(train_generator)
        np.testing.assert_array_equal(ids,expected_ids[epoch][0])
        #batch 1
        ids, X, _ = next(train_generator)
        np.testing.assert_array_equal(ids,expected_ids[epoch][1])
        #batch 2
        ids, X, _ = next(train_generator)
        np.testing.assert_array_equal(ids,expected_ids[epoch][2])
           
    h5.close()


def test_refresh_on_epoch_end_annot():
    """ Test if the correct annotation labels are when the batches are refreshed
    """
    h5 = open_file(os.path.join(path_to_assets, "mini_narw.h5"), 'r') # create the database handle  
    train_data = open_table(h5, "/train/data")
    
    np.random.seed(100)

    def transform_output(x,y):
        X = x
        print(y)
        Y = np.array([(value[0], value[1]) for value in y])
        return X,Y

    ids_in_db = train_data[:]['id']
    train_generator = BatchGenerator(data_table=train_data, batch_size=6,
                                     y_field=['label'], return_batch_ids=True, shuffle=True,
                                     refresh_on_epoch_end=True, output_transform_func=None) #create a batch generator 

    expected_ids = {'epoch_1': ([17, 19, 11, 18, 13,  6], [16,  1,  9, 14, 12,  5], [2,  4, 10,  0, 15, 7,  3,  8]),    
                     'epoch_2':  ([18, 19, 17, 0, 8, 6], [14, 7, 11, 10, 15, 3], [5, 13, 1, 4, 12, 2, 9, 16]),
                     'epoch_3': ([3, 4, 12, 17, 10, 1], [19, 5, 11, 8, 0, 18], [6, 13, 7, 15, 16, 14, 2, 9])}
                     
    expected_labels = {'epoch_1':  ([0, 0, 0, 0, 0, 1], [0, 1, 1, 0, 0, 1], [1, 1, 0, 1, 0, 1, 1, 1]),
                     'epoch_2': ([0, 0, 0, 1, 1, 1], [0, 1, 0, 0, 0, 1], [1, 0, 1, 1, 0, 1, 1, 0]),    
                     'epoch_3': ([1, 1, 0, 0, 0, 1], [0, 1, 0, 1, 1, 0], [1, 0, 1, 0, 0, 0, 1, 1])}
                     
    for epoch in ['epoch_1', 'epoch_2', 'epoch_3']:
        #batch 0
        ids, X, Y = next(train_generator)
        np.testing.assert_array_equal(ids,expected_ids[epoch][0])
        np.testing.assert_array_equal(Y,expected_labels[epoch][0])
        #batch 1
        ids, X, Y = next(train_generator)    
        np.testing.assert_array_equal(ids,expected_ids[epoch][1])
        np.testing.assert_array_equal(Y,expected_labels[epoch][1])
        #batch 2
        ids, X, Y = next(train_generator)
        np.testing.assert_array_equal(ids,expected_ids[epoch][2])
        np.testing.assert_array_equal(Y,expected_labels[epoch][2])
    
    h5.close()


def test_output_transform_function():
    """ Test if the function passed as 'instance_function' is applied to the batch
    """
    h5 = open_file(os.path.join(path_to_assets, "mini_narw.h5"), 'r') # create the database handle  
    train_data = open_table(h5, "/train/data")

    def apply_to_batch(X,Y):
        X = np.mean(X, axis=(1,2))
        return (X, Y)

    train_generator = BatchGenerator(data_table=train_data,  batch_size=6, return_batch_ids=True, output_transform_func=apply_to_batch) #create a batch generator 
    
    _, X, Y = next(train_generator)
    assert X.shape == (6,)
    assert X[0] == pytest.approx(-37.345703, 0.1)
    assert Y.shape == (6,)
    
    h5.close()

def test_extended_batches():
    """ Test that batches can be extended to include last/first samples from previous/next batch
    """
    h5 = open_file(os.path.join(path_to_assets, "mini_narw.h5"), 'r') # create the database handle  
    train_data = open_table(h5, "/train/data")

    ids_in_db = train_data[:]['id']
    train_generator = BatchGenerator(data_table=train_data, batch_size=6, return_batch_ids=True, n_extend=2) #create a batch generator 
    
    #First batch
    ids, X, _ = next(train_generator)
    np.testing.assert_array_equal(ids,[0,1,2,3,4,5,6,7])
    assert X.shape == (8, 94, 129)

    #Second batch
    ids, X, _ = next(train_generator)
    np.testing.assert_array_equal(ids,[4,5,6,7,8,9,10,11,12,13])
    assert X.shape == (10, 94, 129)

    #Third batch; Last batch ( will have the remaining instances)
    ids, X, _ = next(train_generator)
    np.testing.assert_array_equal(ids,[10, 11, 12, 13, 14, 15, 16, 17, 18, 19])
    assert X.shape == (10, 94, 129)
    
    h5.close()

def test_batch_size_larger_than_dataset_size():
    """ Test that batch size can exceed dataset size
    """
    h5 = open_file(os.path.join(path_to_assets, "mini_narw.h5"), 'r') # create the database handle  
    train_data = open_table(h5, "/train/data")

    ids_in_db = train_data[:]['id']
    with warnings.catch_warnings(record=True) as w:
        train_generator = BatchGenerator(data_table=train_data, batch_size=99, return_batch_ids=True) #create a batch generator 

        assert train_generator.batch_size == 20
        assert len(w) == 1
        assert "The batch size is greater than the number of instances available. Setting batch_size to n_instances." in str(w[-1].message)
    
    #First batch
    ids, X, _ = next(train_generator)
    np.testing.assert_array_equal(ids,[0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19])
    assert X.shape == (20, 94, 129)
    
    h5.close()


def test_joint_batch_gen():
    """ Test the a joint batch generator can be used to load from tables with a single data column
    """
    h51 = open_file(os.path.join(path_to_assets, "mini_narw.h5"), 'r') # create the database handle  
    h52 = open_file(os.path.join(path_to_assets, "mini_narw.h5"), 'r') # create the database handle  
    tbl1 = open_table(h51, "/train/data")
    tbl2 = open_table(h52, "/train/data")

    three_specs = tbl1.col('data')[:3]
    three_labels = tbl1.col('label')[:3]

    three_labels = [np.array(l) for l in three_labels]

    gen1 = BatchGenerator(data_table=tbl1, batch_size=3, return_batch_ids=True)  
    gen2 = BatchGenerator(data_table=tbl2, batch_size=2, return_batch_ids=False)  

    gen = JointBatchGen([gen1, gen2], n_batches="min") 
    X, Y = next(gen)
    
    assert len(X) == 5
    assert X[0].shape == (94, 129)
    np.testing.assert_array_equal(X[:3], three_specs)
    np.testing.assert_array_equal(X[3:], three_specs[:2])

    assert len(Y) == 5
    for i in range(3):
        np.testing.assert_array_equal(Y[i], three_labels[i])
    for i in range(2):
        np.testing.assert_array_equal(Y[3+i], three_labels[i])

    h51.close()
    h52.close()

def test_joint_batch_gen_ids():
    """ Test the a joint batch generator can return ids
    """
    h51 = open_file(os.path.join(path_to_assets, "mini_narw.h5"), 'r') # create the database handle  
    h52 = open_file(os.path.join(path_to_assets, "mini_narw.h5"), 'r') # create the database handle  
    tbl1 = open_table(h51, "/train/data")
    tbl2 = open_table(h52, "/train/data")

    three_specs = tbl1.col('data')[:3]
    three_labels = tbl1.col('label')[:3]

    three_labels = [np.array(l) for l in three_labels]

    gen1 = BatchGenerator(data_table=tbl1, batch_size=3, return_batch_ids=True)  
    gen2 = BatchGenerator(data_table=tbl2, batch_size=2, return_batch_ids=False)  

    gen = JointBatchGen([gen1, gen2], n_batches="min", return_batch_ids=True) 
    ids, X, Y = next(gen)
    
    assert len(ids) == 5
    assert len(ids[0]) == 2
    assert np.all(ids[0] == [0, 0])
    assert np.all(ids[1] == [0, 1])
    assert np.all(ids[2] == [0, 2])
    assert np.all(ids[3] == [1, 0])
    assert np.all(ids[4] == [1, 1])

    assert len(X) == 5
    assert X[0].shape == (94, 129)
    np.testing.assert_array_equal(X[:3], three_specs)
    np.testing.assert_array_equal(X[3:], three_specs[:2])

    assert len(Y) == 5
    for i in range(3):
        np.testing.assert_array_equal(Y[i], three_labels[i])
    for i in range(2):
        np.testing.assert_array_equal(Y[3+i], three_labels[i])

    h51.close()
    h52.close()

def test_joint_batch_gen_output_transform():
    """ Test the a joint batch generator can be used to load from tables with a single data column
        while applying the ResNet output transform
    """
    h51 = open_file(os.path.join(path_to_assets, "mini_narw.h5"), 'r') # create the database handle  
    h52 = open_file(os.path.join(path_to_assets, "mini_narw.h5"), 'r') # create the database handle  
    tbl1 = open_table(h51, "/train/data")
    tbl2 = open_table(h52, "/train/data")

    three_specs = tbl1.col('data')[:3]
    three_labels = tbl1.col('label')[:3]

    three_labels = [np.array(l) for l in three_labels]

    gen1 = BatchGenerator(data_table=tbl1, batch_size=3, return_batch_ids=True, output_transform_func=ResNetInterface.transform_batch)  
    gen2 = BatchGenerator(data_table=tbl2, batch_size=2, return_batch_ids=False, output_transform_func=ResNetInterface.transform_batch)  

    gen = JointBatchGen([gen1, gen2], n_batches="min") 
    X, Y = next(gen)
    
    assert len(X) == 5
    assert X[0].shape == (94, 129, 1)
    np.testing.assert_array_equal(X[:3,:,:,0], three_specs)
    np.testing.assert_array_equal(X[3:,:,:,0], three_specs[:2])

    assert len(Y) == 5
    np.testing.assert_array_equal(Y[:3], np.array([[0,1],[0,1],[0,1]]))
    np.testing.assert_array_equal(Y[3:], np.array([[0,1],[0,1]]))

    h51.close()
    h52.close()

def test_joint_batch_gen_multi_modal():
    """ Test the a joint batch generator can be used to load multi-modal data
    """
    h51 = open_file(os.path.join(path_to_assets, "mini_narw_mult.h5"), 'r') # create the database handle  
    h52 = open_file(os.path.join(path_to_assets, "mini_narw_mult.h5"), 'r') # create the database handle  
    tbl1 = open_table(h51, "/train/data")
    tbl2 = open_table(h52, "/train/data")

    three_specs = tbl1.col('spec')[:3]
    three_gammas = tbl1.col('gamma')[:3]
    three_labels = tbl1.col('label')[:3]

    three_labels = [np.array(l) for l in three_labels]

    gen1 = BatchGenerator(data_table=tbl1, batch_size=3, x_field=['spec','gamma'])  
    gen2 = BatchGenerator(data_table=tbl2, batch_size=2, x_field=['spec','gamma'])  

    gen = JointBatchGen([gen1, gen2], n_batches="min", return_batch_ids=True) 
    ids, X, Y = next(gen)

    assert len(ids) == 5
    assert len(ids[0]) == 2
    assert np.all(ids[0] == [0, 0])
    assert np.all(ids[1] == [0, 1])
    assert np.all(ids[2] == [0, 2])
    assert np.all(ids[3] == [1, 0])
    assert np.all(ids[4] == [1, 1])

    assert len(X) == 5
    assert len(X[0]) == 2
    assert X[0]['spec'].shape == (94, 129)
    assert X[0]['gamma'].shape == (3000, 20)
    specs = [x[0] for x in X]
    gammas = [x[1] for x in X]
    np.testing.assert_array_equal(specs[:3], three_specs)
    np.testing.assert_array_equal(gammas[:3], three_gammas)
    np.testing.assert_array_equal(specs[3:], three_specs[:2])
    np.testing.assert_array_equal(gammas[3:], three_gammas[:2])

    assert len(Y) == 5
    labels = [y for y in Y]
    np.testing.assert_array_equal(labels[:3], three_labels)
    np.testing.assert_array_equal(labels[3:], three_labels[:2])

    h51.close()
    h52.close()

def test_joint_batch_gen_multi_modal_transform():
    """ Test the a joint batch generator can be used to load multi-modal data
        while applying output transform """
    h51 = open_file(os.path.join(path_to_assets, "mini_narw_mult.h5"), 'r') # create the database handle  
    h52 = open_file(os.path.join(path_to_assets, "mini_narw_mult.h5"), 'r') # create the database handle  
    tbl1 = open_table(h51, "/train/data")
    tbl2 = open_table(h52, "/train/data")

    three_specs = tbl1.col('spec')[:3]
    three_gammas = tbl1.col('gamma')[:3]
    three_labels = tbl1.col('label')[:3]

    three_labels = [np.array(l) for l in three_labels]

    def transform_batch(X, Y):
        X = [[x['spec'][:,:,np.newaxis], x['gamma'][:,:,np.newaxis]] for x in X]
        Y = np.array([label for label in Y])        
        return (X,Y)

    gen1 = BatchGenerator(data_table=tbl1, batch_size=3, x_field=['spec','gamma'], output_transform_func=transform_batch)  
    gen2 = BatchGenerator(data_table=tbl2, batch_size=2, x_field=['spec','gamma'], output_transform_func=transform_batch)  

    gen = JointBatchGen([gen1, gen2], n_batches="min") 
    X, Y = next(gen)
    
    assert len(X) == 5
    assert len(X[0]) == 2
    assert X[0][0].shape == (94, 129, 1)
    assert X[0][1].shape == (3000, 20, 1)
    specs = [x[0][:,:,0] for x in X]
    gammas = [x[1][:,:,0] for x in X]
    np.testing.assert_array_equal(specs[:3], three_specs)
    np.testing.assert_array_equal(gammas[:3], three_gammas)
    np.testing.assert_array_equal(specs[3:], three_specs[:2])
    np.testing.assert_array_equal(gammas[3:], three_gammas[:2])

    assert len(Y) == 5
    np.testing.assert_array_equal(Y[:3], three_labels)
    np.testing.assert_array_equal(Y[3:], three_labels[:2])

    h51.close()
    h52.close()


def test_multi_modal_batch_generator():
    """ Test that the multi-modal batch generator behaves as expected
    """
    h51 = open_file(os.path.join(path_to_assets, "mini_narw_mult.h5"), 'r') # create the database handle  
    h52 = open_file(os.path.join(path_to_assets, "mini_narw_mult.h5"), 'r') # create the database handle  
    tbl1 = open_table(h51, "/train/data")
    tbl2 = open_table(h52, "/train/data")

    three_specs = tbl1.col('spec')[:3]
    three_gammas = tbl1.col('gamma')[:3]
    three_labels = tbl1.col('label')[:3]

    def transform_batch(X, Y):
        # invert ordering: batch,mode -> mode,batch
        # also, stack the inner (batch) dimension to create numpy arrays
        X = [np.stack([x[i] for x in X]) for i in range(len(X[0]))]
        Y = [np.stack([y[i] for y in Y]) for i in range(len(Y[0]))]
        # only pick the labels from the first mode:
        Y = Y[0]
        return (X,Y)

    gen1 = BatchGenerator(data_table=tbl1, batch_size=3, x_field='spec')  
    gen2 = BatchGenerator(data_table=tbl2, batch_size=3, x_field='gamma')  
    gen = MultiModalBatchGen([gen1, gen2], output_transform_func=transform_batch) 
    X, Y = next(gen)
    assert np.all(X[0] == three_specs)
    assert np.all(X[1] == three_gammas)
    assert np.all(Y == three_labels)

    gen1 = BatchGenerator(data_table=tbl1, batch_size=3, x_field='spec')  
    gen2 = BatchGenerator(data_table=tbl2, batch_size=3, x_field='spec')  
    gen = MultiModalBatchGen([gen1, gen2], shuffle=True, output_transform_func=transform_batch) 
    X, Y = next(gen)
    assert np.all(X[0] == X[1])

    h51.close()
    h52.close()


# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

import pytest
import datetime
import os
from pathlib import Path
import shutil
import tempfile
import numpy as np
import scipy.signal as sg
import soundfile as sf
import pandas as pd
import tarfile
import ketos.audio.utils.misc as ap
from ketos.data_handling.data_handling import to1hot
from ketos.data_handling.data_feeding import BatchGenerator
from ketos.audio.waveform import Waveform
from ketos.audio.utils.axis import LinearAxis, Log2Axis, MelAxis
import ketos.audio.base_audio as aba

path_to_assets = os.path.join(os.path.dirname(__file__),"assets")
path_to_tmp = os.path.join(path_to_assets,'tmp')

if not os.path.exists(path_to_tmp):
    os.makedirs(path_to_tmp)


# This fixture will create a tmp path and remove it after we are done
@pytest.fixture(scope='function')
def tmpdir():
    tmp = tempfile.mkdtemp()
    yield Path(tmp)
    shutil.rmtree(tmp)

@pytest.fixture
def sine_wave():
    sampling_rate = 44100
    frequency = 2000
    duration = 3
    x = np.arange(duration * sampling_rate)
    signal = np.sin(2 * np.pi * frequency * x / sampling_rate) 
    return sampling_rate, signal

@pytest.fixture
def square_wave():
    sampling_rate = 44100
    frequency = 2000
    duration = 3
    x = np.arange(duration * sampling_rate)
    signal = sg.square(2 * np.pi * frequency * x / sampling_rate) 
    return sampling_rate, signal

@pytest.fixture
def sawtooth_wave():
    sampling_rate = 44100
    frequency = 2000
    duration = 3
    x = np.arange(duration * sampling_rate)
    signal = sg.sawtooth(2 * np.pi * frequency * x / sampling_rate) 
    return sampling_rate, signal

@pytest.fixture
def const_wave():
    sampling_rate = 44100
    duration = 3
    x = np.arange(duration * sampling_rate)
    signal = np.ones(len(x))
    return sampling_rate, signal

@pytest.fixture
def growing_sine_wave():
    sampling_rate = 1000
    frequency = 18
    duration = 3
    x = np.arange(duration * sampling_rate)
    signal = x / np.max(x) * np.sin(2 * np.pi * frequency * x / sampling_rate) 
    return sampling_rate, signal

@pytest.fixture
def sine_wave_file(sine_wave):
    """Create a .wav with the 'sine_wave()' fixture
    
       The file is saved as tests/assets/tmp/sine_wave.wav.
       When the tests using this fixture are done, 
       the file is deleted.

       Yields:
            wav_file : str
                A string containing the path to the .wav file.
    """
    wav_file = os.path.join(path_to_tmp, "sine_wave.wav")
    rate, sig = sine_wave
    sf.write(wav_file, sig, rate)    
    yield wav_file
    os.remove(wav_file)

@pytest.fixture
def sine_wave_file_half(sine_wave):
    """Create a .wav with the 'sine_wave()' fixture, with an amplitude
        of 0.5 instead of 1.

       The file is saved as tests/assets/tmp/sine_wave_half.wav.
       When the tests using this fixture are done, 
       the file is deleted.

       Yields:
            wav_file : str
                A string containing the path to the .wav file.
    """
    wav_file = os.path.join(path_to_tmp, "sine_wave_half.wav")
    rate, sig = sine_wave
    sf.write(wav_file, 0.5*sig, rate)    
    yield wav_file
    os.remove(wav_file)

@pytest.fixture
def square_wave_file(square_wave):
    """Create a .wav with the 'square_wave()' fixture
    
       The file is saved as tests/assets/tmp/square_wave.wav.
       When the tests using this fixture are done, 
       the file is deleted.

       Yields:
            wav_file : str
                A string containing the path to the .wav file.
    """
    wav_file =  os.path.join(path_to_tmp, "square_wave.wav")
    rate, sig = square_wave
    sf.write(wav_file, sig, rate)    
    yield wav_file
    os.remove(wav_file)

@pytest.fixture
def sawtooth_wave_file(sawtooth_wave):
    """Create a .wav with the 'sawtooth_wave()' fixture
    
       The file is saved as tests/assets/tmp/sawtooth_wave.wav.
       When the tests using this fixture are done, 
       the file is deleted.

       Yields:
            wav_file : str
                A string containing the path to the .wav file.
    """
    wav_file =  os.path.join(path_to_tmp, "sawtooth_wave.wav")
    rate, sig = sawtooth_wave
    sf.write(wav_file, sig, rate)    
    yield wav_file
    os.remove(wav_file)

@pytest.fixture
def const_wave_file(const_wave):
    """Create a .wav with the 'const_wave()' fixture
    
       The file is saved as tests/assets/tmp/const_wave.wav.
       When the tests using this fixture are done, 
       the file is deleted.

       Yields:
            wav_file : str
                A string containing the path to the .wav file.
    """
    wav_file =  os.path.join(path_to_tmp, "const_wave.wav")
    rate, sig = const_wave
    sf.write(wav_file, sig, rate)    
    yield wav_file
    os.remove(wav_file)

@pytest.fixture
def growing_sine_wave_file(growing_sine_wave):
    """Create a .wav with the 'growing_sine_wave()' fixture
    
       The file is saved as tests/assets/tmp/growing_sine_wave.wav.
       When the tests using this fixture are done, 
       the file is deleted.

       Yields:
            wav_file : str
                A string containing the path to the .wav file.
    """
    wav_file = os.path.join(path_to_tmp, "growing_sine_wave.wav")
    rate, sig = growing_sine_wave
    sf.write(wav_file, sig, rate)    
    yield wav_file
    os.remove(wav_file)

@pytest.fixture
def image_2x2():
    image = np.array([[1,2],[3,4]], np.float32)
    return image

@pytest.fixture
def image_3x3():
    image = np.array([[1,2,3],[4,5,6],[7,8,9]], np.float32)
    return image

@pytest.fixture
def image_ones_10x10():
    image = np.ones(shape=(10,10))
    return image

@pytest.fixture
def image_zeros_and_ones_10x10():
    image = np.ones(shape=(10,10))
    for i in range(10):
        for j in range(5):
            image[i,j] = 0
    return image

@pytest.fixture
def datebase_with_one_image_col_and_one_label_col(image_2x2):
    d = {'image': [image_2x2], 'label': [1]}
    df = pd.DataFrame(data=d)
    return df


@pytest.fixture
def datebase_with_one_image_col_and_no_label_col(image_2x2):
    d = {'image': [image_2x2]}
    df = pd.DataFrame(data=d)
    return df


@pytest.fixture
def datebase_with_two_image_cols_and_one_label_col(image_2x2):
    d = {'image1': [image_2x2], 'image2': [image_2x2], 'label': [1]}
    df = pd.DataFrame(data=d)
    return df


@pytest.fixture
def database_prepared_for_NN(image_2x2):
    d = {'image': [image_2x2, image_2x2, image_2x2, image_2x2, image_2x2, image_2x2], 'label': [0,0,0,0,0,0]}
    df = pd.DataFrame(data=d)
    divisions = {"train":(0,3),"validation":(3,4),"test":(4,6)}
    prepared = prepare_database(df, "image", "label", divisions)     
    return prepared

@pytest.fixture
def database_prepared_for_NN_2_classes():
    img1 = np.zeros((20, 20))
    img2 = np.ones((20, 20))
    d = {'image': [img1, img2, img1, img2, img1, img2,
                   img1, img2, img1, img2, img1, img2,
                   img1, img2, img1, img2, img1, img2,
                   img1, img2, img1, img2, img1, img2],
         'label': [0, 1, 0, 1, 0, 1,
                   0, 1, 0, 1, 0, 1,
                   0, 1, 0, 1, 0, 1,
                   0, 1, 0, 1, 0, 1]}
    database = pd.DataFrame(data=d)
    divisions= {"train":(0,12),
                "validation":(12,18),
                "test":(18,len(database))}
    prepared = prepare_database(database=database,x_column="image",y_column="label",
                                divisions=divisions)    
    return prepared


@pytest.fixture
def sine_audio(sine_wave):
    rate, data = sine_wave
    a = Waveform(rate=rate, data=data, filename='sine_wave')
    return a
    
@pytest.fixture
def data_classified_by_nn():
    x = [1, 2, 3, 4, 5, 6] # input data
    x = np.array(x)
    y = [0, 1, 0, 1, 0, 1] # labels
    y = np.array(y)
    w = [[0.8, 0.2], [0.1, 0.9], [0.96, 0.04], [0.49, 0.51], [0.45, 0.55], [0.60, 0.40]] # class weights computed by NN
    w = np.array(w)
    return x,y,w

@pytest.fixture
def data_for_TCN():
    fv0 = np.zeros(64)
    fv1 = np.ones(64)
    x_train = np.array([fv0, fv1, fv0, fv1, fv0, fv1, fv0, fv1, fv0, fv1])
    y_train = np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 1])
    x_val = np.array([fv0, fv1, fv0, fv1])
    y_val = np.array([0, 1, 0, 1])
    x_test = np.array([fv0, fv1, fv0, fv1])
    y_test = np.array([0, 1, 0, 1])
    return x_train, y_train, x_val, y_val, x_test, y_test
    

def encode_database(database, x_column, y_column):
    image_shape = database[x_column][0].shape
    depth = database[y_column].max() + 1 #number of classes
    database["one_hot_encoding"] = database[y_column].apply(to1hot,depth=depth)
    database["x_flatten"] = database[x_column].apply(lambda x: x.flatten())
    return database, image_shape

def split_database(database, divisions):
    train_data = database[divisions["train"][0]:divisions["train"][1]]
    validation_data = database[divisions["validation"][0]:divisions["validation"][1]]
    test_data = database[divisions["test"][0]:divisions["test"][1]]
    datasets = {"train": train_data,
                "validation": validation_data,
                "test": test_data}
    return datasets

def stack_dataset(dataset, input_shape):
    x = np.vstack(dataset.x_flatten).reshape(dataset.shape[0], input_shape[0], input_shape[1],1).astype(np.float32)
    y = np.vstack(dataset.one_hot_encoding)
    stacked_dataset = {'x': x,
                       'y': y}
    return stacked_dataset

def prepare_database(database, x_column, y_column, divisions):
    encoded_data, input_shape = encode_database(database=database, x_column=x_column, y_column=y_column)
    datasets = split_database(database=encoded_data, divisions=divisions)
    stacked_train = stack_dataset(dataset=datasets["train"], input_shape=input_shape)
    stacked_validation = stack_dataset(dataset=datasets["validation"], input_shape=input_shape)
    stacked_test = stack_dataset(dataset=datasets["test"], input_shape=input_shape)
    stacked_datasets = {"train_x": stacked_train["x"],
                        "train_y": stacked_train["y"],
                        "validation_x": stacked_validation["x"],
                        "validation_y": stacked_validation["y"],
                        "test_x": stacked_test["x"],
                        "test_y": stacked_test["y"]}
    return stacked_datasets


@pytest.fixture
def file_duration_table():
    """ Create a table of file durations as a pandas DataFrame.

        Yields:
            tbl: pandas DataFrame
                File duration table
    """
    N = 6
    filename = ['f{0}.wav'.format(x) for x in np.arange(N)]
    duration = [x + 30.0 for x in np.arange(N)]
    tbl = pd.DataFrame({'filename': filename, 'duration': duration})
    return tbl

########################## The following fixtures were created for the selection_table tests. 
@pytest.fixture
def basic_dataframe():
    return pd.DataFrame({
        'filename': ['file1', 'file2'],
        'label': ['whale', 'dolphin'],
        'start': [0, 5],
        'end': [10, 15]
    })

@pytest.fixture
def multi_label_dataframe():
    return pd.DataFrame({
        'filename': ['file1'],
        'label': ['whale,dolphin'],
        'start': [0],
        'end': [10]
    })

@pytest.fixture
def datetime_dataframe():
    return pd.DataFrame({
        'filename': ['file1_20_01_2020'],
        'label': ['whale'],
        'datetime': ['20_01_2020']
    })

@pytest.fixture
def varied_labels_dataframe():
    return pd.DataFrame({
        'filename': ['file1', 'file2', 'file3'],
        'label': ['whale', 'dolphin', 'shark'],
        'start': [0, 5, 10],
        'end': [10, 15, 20]
    })

# Fixture for creating a temporary CSV file
@pytest.fixture
def temp_csv_file():
    data = """filename,label,start,end
            file1,whale,0,10
            file2,dolphin,5,15"""
    with tempfile.NamedTemporaryFile(mode='w+', suffix='.csv', delete=False) as tmpfile:
        tmpfile.write(data)
        return tmpfile.name

@pytest.fixture
def annot_table_std():
    """ Create a standardized annotations table as a pandas DataFrame.

        Yields:
            tbl: pandas DataFrame
                Annotation table
    """
    label = [1, 2, 3, 0, 0, -1]
    N = len(label)
    filename = ['f{0}.wav'.format(x%3) for x in np.arange(N)]
    start = np.arange(N, dtype=float)
    end = start + 3.3
    tbl = pd.DataFrame({'filename': filename, 'label': label, 'start': start, 'end': end})
    return tbl

##############################################

@pytest.fixture
def annot_table():
    """ Create an annotations table as a pandas DataFrame.

        Yields:
            tbl: pandas DataFrame
                Annotation table
    """
    label = [1, 2, 'k', -99, 'whale', 'zebra']
    N = len(label)
    filename = ['f{0}.wav'.format(x) for x in np.arange(N)]
    start = np.arange(N)
    stop = start + 1
    tbl = pd.DataFrame({'fname': filename, 'label': label, 'start': start, 'STOP': stop})
    return tbl

@pytest.fixture
def annot_table_mult_labels():
    """ Create an annotations table as a pandas DataFrame with 
        multiple labels per row.

        Yields:
            tbl: pandas DataFrame
                Annotation table
    """
    label = ['1,2', 3]
    N = len(label)
    filename = ['f{0}.wav'.format(x) for x in np.arange(N)]
    start = np.arange(N)
    end = start + 1
    tbl = pd.DataFrame({'filename': filename, 'label': label, 'start': start, 'end': end})
    return tbl

@pytest.fixture
def annot_table_file(annot_table):
    """ Create an annotations table csv file with the 'annot_table()' fixture
    
        The file is saved as tests/assets/tmp/annot_002.csv.
        When the tests using this fixture are done, 
        the file is deleted.

        Yields:
            csv_file : str
                A string containing the path to the .csv file.
    """
    csv_file = os.path.join(os.path.join(path_to_assets, "tmp"), "annot_002.csv")
    tbl = annot_table
    tbl.to_csv(csv_file, index=False)
    yield csv_file
    os.remove(csv_file)

@pytest.fixture
def linear_axis_200():
    """ Create a linear axis with range 0-100 and 200 bins.

        Yields:
            ax: LinearAxis
                Axis object
    """
    ax = LinearAxis(bins=200, extent=(0.,100.))
    return ax

@pytest.fixture
def log2_axis_8_16():
    """ Create a log2 axis with 8 octaves, 16 bins per octave, and 
        minimum value of 10.

        Yields:
            ax: Log2Axis
                Axis object
    """
    ax = Log2Axis(bins=8*16, bins_per_oct=16, min_value=10.)
    return ax

@pytest.fixture
def mel_axis_40_500():
    """ Create a Mel frequency axis with with 40 bins (filters)
        for a maximum frequncy of 500 Hz

        Yields:
            ax: MelAxis
                Axis object
    """
    ax = MelAxis(num_filters=40, freq_max=500)
    return ax

@pytest.fixture
def spec_image_with_attrs():
    """ Creates a spectrogram image with shape (20,10) and random pixel values, 
        with time resolution of 0.5 s, and a linear frequency axis from 0 to 
        500 Hz.

        Yields:
            img: 2d numpy array
                Pixel values
            dt: float
                Time resolution
            ax: LinearAxis
                Frequency axis            
    """
    img = np.random.rand(20,10)
    dt = 0.5
    ax = LinearAxis(bins=10, extent=(0.,500.), label='Frequency (Hz)')
    return img,dt,ax

@pytest.fixture
def base_audio_1d():
    """ Create a simple 1d BaseAudio object with value 1 everywhere, length 
        of 10 s, filename 'x', offset of 2 s, and label 13.

        Yields:
            o: BaseAudio
                BaseAudio object
    """
    N = 10000
    d = np.ones(N)
    o = aba.BaseAudio(data=d, filename='x', offset=2., label=13, duration=10.)
    return o, d

@pytest.fixture
def base_audio_time_1d():
    """ Create a simple 1d BaseAudioTime object with value 1 everywhere, length 
        of 10 s, time resolution of 0.001 s, filename 'x', offset of 2 s, and label 13.

        Yields:
            o: BaseAudio
                BaseAudio object
    """
    N = 10000
    d = np.ones(N)
    o = aba.BaseAudioTime(time_res=0.001, data=d, filename='x', offset=2., label=13)
    return o, d

@pytest.fixture
def five_time_stamped_wave_files():
    """ Each wav file is 0.5 second long. The sampling rate is 1 kHz."""
    N = 5
    path_to_tmp = os.path.join(path_to_assets,'tmp')
    folder = path_to_tmp + '/five_time_stamped_wave_files/'
    if not os.path.exists(folder):
        os.makedirs(folder)

    files = list()
    for i in range(N):
        fname = 'empty_HMS_12_ 5_ {0}__DMY_23_ 2_84.wav'.format(i)
        full_path = os.path.join(folder, fname)
        a = Waveform(rate=1000, data=np.random.uniform(size=500))
        a.to_wav(full_path)
        files.append(full_path)

    yield folder

    for f in files:
        os.remove(f)

@pytest.fixture
def spectr_settings():
    j = '{"spectrogram": {"type":"MagSpectrogram", "rate": "20 kHz",\
        "window": "0.1 s", "step": "0.025 s",\
        "window_func": "hamming", "freq_min": "30Hz", "freq_max": "3000Hz",\
        "duration": "1.0s", "resample_method": "scipy", "normalize_wav": "False", "decibel": "true",\
        "transforms": [{"name":"enhance_signal", "enhancement":1.0}, {"name":"adjust_range", "range":"(0,1)"}],\
        "waveform_transforms": [{"name":"add_gaussian_noise", "sigma":0.2}] }}'
    return j

@pytest.fixture
def custom_audio_representation_module():
    """Creates a .py file with a custom audio representation class
    
        The file is saved as tests/assets/custom_representation.py.
        When the tests using this fixure are dones, the file is deleted.

        Yields:
            python_file : str
                A string containing the path to the .py file.
    """
    python_file =  os.path.join(path_to_assets, "custom_representation.py")
    class_string = "class CustomRepresentation():\n\
        def __init__(self):\n\
            self.window = '0.2'"
    file = open(python_file, "w+")
    file.write(class_string)
    file.close()
    yield python_file
    os.remove(python_file)

@pytest.fixture
def sample_data():
    data = np.vstack([np.zeros((10,64,64,1)), np.ones((10,64,64,1))])
    labels = np.concatenate([np.array([[1,0] for i in range(10)]), np.array([[0,1] for i in range(10)])])
    return (data, labels)

@pytest.fixture
def sample_data_1d():
    data = np.vstack([np.zeros((10,20000,1)), np.ones((10,20000,1))])
    labels = np.concatenate([np.array([[1,0] for i in range(10)]), np.array([[0,1] for i in range(10)])])
    return (data, labels)

@pytest.fixture
def tar_archive_with_wav_files():
    """Create a .tar file containing several wav files.
    
       The file is saved as tests/assets/audio_archive.tar.
       When the tests using this fixture are done, 
       the file is deleted.

       Yields:
            tar_path : str
                A string containing the path to the .tar file.
    """
    tar_path = os.path.join(path_to_assets, "audio_archive.tar")
    tar = tarfile.open(tar_path , "w") 
    tar.add(os.path.join(path_to_assets, os.path.join('wav_files','w1.wav')), arcname="w1.wav")
    tar.add(os.path.join(path_to_assets, os.path.join('wav_files','w1.wav')), arcname=os.path.join("a","w2.wav"))        
    tar.add(os.path.join(path_to_assets, os.path.join('wav_files','w1.wav')), arcname=os.path.join("a","w3.wav"))        
    tar.add(os.path.join(path_to_assets, os.path.join('wav_files','w1.wav')), arcname=os.path.join("a",os.path.join("b","w1.wav")))        
    tar.close()
    yield tar_path
    os.remove(tar_path)


import pickle
import pandas as pd
import os
import yaml
import numpy as np
import sys
import warnings
from ketos.ketos_commands.constants import EXIT_ERROR, EXIT_INTERRUPTION, EXIT_SUCCESS
from pathlib import Path
from ketos.ketos_commands.ketos_logger import KetosLogger


def _ketos_run_output_function(batch_detections, output_folder, mode, threshold=0.5, merge_detections=False, buffer=None, running_avg=None, labels=None, highest_score_only=False, **output_function_arguments):
    """ Converts network raw output into detection data.

        This function processes the output of a network into user-friendly detection data, 
        which is then saved to a .csv file. By default, each detection instance is stored 
        as a separate line in the .csv file.

        While this function offers a default conversion method, it can be overwritten to suit 
        specific use cases, transforming the network output based on user requirements.

        The contents of the network output batch (batch_detections) will 
        vary depending on the type of network. The format of batch_detections for various types of classifiers are as follows:

        Binary Classifier:

        {
         'filename': ['file1.wav', 'file1.wav', 'file1.wav'], 
         'start': [0.0, 3.0, 6.0], 
         'end': [3.0, 6.0, 9.0], 
         'score': [[0.3, 0.7], [0.8, 0.2], [0.1, 0.9]]
        }

        Multiclass Classifier:

        {
         'filename': ['file1.wav', 'file1.wav', 'file1.wav'], 
         'start': [0.0, 3.0, 6.0], 
         'end': [3.0, 6.0, 9.0], 
         'score': [[0.3, 0.6, 0.1], [0.6, 0.2, 0.2], [0.3, 0.4, 0.3]]
        }

        Sequence Classifier:

        {
         'filename': ['file1.wav'],
         'start': [0.0],
         'end': [10.0],
         'score': [[
             [0.8, 0.9, 0.7, 0.2, 0.1, 0.6, 0.9, 0.9, 0.2, 0.6],
             [0.7, 0.4, 0.6, 0.8, 0.6, 0.7, 0.8, 0.6, 0.1, 0.5]
         ]]
        }

        Args:
            batch_detections: dict
                A dictionary containing network outputs, including "filename", "start", "end", and "score" fields.
            output_folder: str
                The path to the folder where the output .csv file will be saved.
            mode: str
                The file I/O mode to be used while writing the .csv file. Typically 'w' for write or 'a' for append.
            threshold: float
                The score threshold for filtering detections. Defaults to 0.5.
            merge_detections: bool
                If True, overlapping detections will be merged into single detections. Defaults to False.
            buffer: float
                Adds buffer (seconds) to the start and end times of each detection. Defaults to None.
            running_avg: float
                If set, computes the running average of scores over the provided window. Defaults to None.
            labels: list or integer 
                A list of labels to filter by. Defaults to None
            highest_score_only: boolean
                If True, will only return the label associated with the highest score even if more than one passes the threshold. Defaults to False.
            output_function_arguments: dict
                Additional keyword arguments for the function.
    """    
    from ketos.neural_networks.dev_utils.detection import convert_sequence_to_snapshot, add_detection_buffer, merge_overlapping_detections, filter_by_threshold, filter_by_label, compute_score_running_avg

    if running_avg is not None:
        batch_detections['score'] = compute_score_running_avg(batch_detections['score'], running_avg)
    

    if batch_detections['score'].ndim == 2: # snapshot
        batch_detections = filter_by_threshold(batch_detections, threshold, highest_score_only)
    else: # sequence model
        batch_detections = convert_sequence_to_snapshot(batch_detections, threshold)

    if labels is not None:
        batch_detections = filter_by_label(batch_detections, labels)

    if merge_detections:
        batch_detections = merge_overlapping_detections(batch_detections)

    if buffer is not None:
        batch_detections = add_detection_buffer(batch_detections, buffer)

    header = True if mode == "w" else False
    output = (output_folder / "detections.csv")

    if not batch_detections.empty:
        batch_detections.to_csv(output, mode=mode, index=False, header=header)
    elif mode == 'w' or not output.exists():
        # If batch_detections is empty and mode is 'w', write an empty DataFrame to the file.
        empty_df = pd.DataFrame(columns=['filename', 'start', 'end', 'label', 'score'])
        empty_df.to_csv(output, mode='w', index=False, header=header)


def get_last_processed_batch(log_file):
    """
    Reads the log file, extracts the last processed batch number and returns it.

    This function is designed to recover the state of a process in case of an
    interruption or crash. It opens the log file and reads the log entries
    in reverse order, looking for the most recent 'Step' entry. It then 
    parses this entry to extract the batch number, which it returns.

    The function also handles two special cases: 
    1. If the log file does not exist (indicating that no batches have been 
       processed), it returns -1.
    2. If the last line in the log file is 'Process finished successfully', 
       it returns None. This is to indicate that the process was completed 
       successfully and does not need to be resumed.

    Parameters:
        log_file (str): Path to the log file.

    Returns:
        int, None: The number of the last processed batch, or None if the 
               process finished successfully.
    """

    last_processed_batch = -1
    try:
        with open(log_file, 'r') as f:
            docs = yaml.load_all(f, Loader=yaml.SafeLoader)

            for doc in docs:
                if 'step' in doc:
                    last_processed_batch = doc.get('step')
                elif doc.get('event').startswith('Process finished.'):
                    last_processed_batch = None

    except FileNotFoundError:
        # Log file doesn't exist, no batches have been processed
        warnings.warn(
            "Overwrite is set to False, but no existing log file was found to resume processing. The process will start from the beginning.",
            category=UserWarning
        )
        return -1
    return last_processed_batch

def _get_num_batches_from_hdf5(hdf5_file_path, batch_size, table_name=None):
    """ Helper function to return the number of batches in an HDF5 file."""
    import tables
    if table_name is None:
        table_name = "/data"
    else:
        table_name = table_name + "/data"

    with tables.open_file(hdf5_file_path, 'r') as hdf5_file:
        table = hdf5_file.get_node(table_name)
        num_samples = table.shape[0]
        num_batches = int(np.ceil(num_samples / batch_size))

    return num_batches 

def ketos_run(model_file, audio_data, file_list=None, table_name=None, output_folder=None, overwrite=True, log_file="ketos-run.log", step_size=None, batch_size=32, threshold=0.5, labels=None, merge_detections=False, buffer=None, running_avg=None, highest_score_only=False, output_function_arguments=None):
    """ 
    Executes a model on given audio data and logs the process.

    This function handles loading and running a machine learning model trained
    with Ketos. It can operate on either a single audio file, a list of files,
    or an entire directory. The function logs its progress, including which
    audio files have been processed, and can resume from where it left off
    in the event of an interruption.

    Args:
        model_file: str
            Path to the .kt file containing the trained model.
        audio_data: str
            Path to the audio data to be processed. This can be a path to a single audio file, a directory of audio files, or an HDF5 database.
        file_list: str
            Path to a CSV file containing a list of specific files to be processed. Defaults to None.
        table_name: str
            Name of the table in the hdf5 file, if applicable. Defaults to None.
        output_folder: str
            Directory where the output will be saved. If not specified, defaults to current directory.
        overwrite: bool
            Whether to overwrite existing files. Defaults to True.
        log_file: str
            Name of the log file. Defaults to "ketos-run.log".
        step_size: float
            Size of the step for the audio frame loader. If not specified, set to the duration of the audio representation.
        batch_size: int
            Size of the data batches that the model will process. Defaults to 32.
        threshold: float
            Threshold for classifying detections. Defaults to 0.5.
        labels: list or integer 
                A list of labels to filter by. Defaults to None
        merge_detections: bool
            Whether to merge detections in output. Defaults to False.
        buffer: int
            Buffer to be added to start and end of each detection. Defaults to None.
        running_avg: float
            Exponential decay factor to compute the running average of detections. Defaults to None.
        highest_score_only: boolean
            If True, will only return the label associated with the highest score even if more than one passes the threshold. Defaults to False.
        output_function_arguments: dict
            Additional arguments to be passed to the output function. Defaults to None.
    """

    from ketos.audio.audio_loader import AudioFrameLoader
    from ketos.audio import load_audio_representation_from_file
    from ketos.neural_networks import load_model_file, load_input_transform_function, load_export_transform_function
    from ketos.neural_networks.dev_utils.detection import batch_load_audio_file_data, batch_load_hdf5_data

    if output_function_arguments is None:
        output_function_arguments = {}

    output_function_arguments.update({
        'threshold': threshold,
        'labels': labels,
        'merge_detections': merge_detections,
        'buffer': buffer,
        'running_avg': running_avg,
        # add any other optional arguments here
    })

    if output_folder is None:
        output_folder = Path('.').resolve()
    else:
        output_folder = Path(output_folder).resolve()
    
    output_folder.parent.mkdir(parents=True, exist_ok=True)

    mode = "w"
    last_processed_batch = -1
    if not overwrite:
        mode = "a"

        log_file_path = output_folder / log_file
        last_processed_batch = get_last_processed_batch(log_file_path)
        if last_processed_batch is None:
            print("There are no more files to process.")
            return

    # load the model from the .kt file
    model = load_model_file(model_file, load_audio_repr=False) 
    
    # load the audio representation from the .kt file
    audio_representation = load_audio_representation_from_file(model_file)

    # Currently the script only works with one audio representation
    audio_representation_name, audio_representation = next(iter(audio_representation[0].items()))

    # Try to load input transform function from model file. Otherwise try to load the one defined on the model, otherwise there is none
    try:
        input_transform_func = load_input_transform_function(model_file)
        # Warn the user about using a custom input transform function
        warnings.warn("Ketos detected a custom input function and will be using that.", UserWarning)
    except AttributeError:
        try:
            input_transform_func = model.transform_batch
        except:
            input_transform_func = None
    # if step size was not specified, set it equal to the window size
    if step_size is None:
        step_size = audio_representation['duration']

    output = (output_folder / "raw_output.pkl")

    
    #  try to load from the model file. Otherwise load the default one
    try:
        output_function = load_export_transform_function(model_file)
        # Warn the user about using a custom input transform function
        warnings.warn("Ketos detected a custom output function and will be using that.", UserWarning)
    except AttributeError:
        output_function = _ketos_run_output_function
    
    audio_path = Path(audio_data)

    # Check if audio_data is a folder (thereby containing wav files) or an hdf5 database
    if audio_path.is_file():
        if audio_path.suffix == '.h5':
            # is hdf5
            batch_generator = lambda: batch_load_hdf5_data(hdf5_file_path=audio_path, batch_size=batch_size, audio_representation=audio_representation, start_idx=last_processed_batch+1, table_name=table_name, x_field=audio_representation_name, logger=logger)
            
            n_steps = _get_num_batches_from_hdf5(hdf5_file_path=audio_path, batch_size=batch_size, table_name=table_name)
    
        elif audio_path.suffix == '.wav':
            # is a single audio file
            audio_files = [audio_path]

            # initialize audio loader for a single file
            loader = AudioFrameLoader(duration=audio_representation["duration"], pad=False, step=step_size, batch_size=1,
                                    filename=audio_files, representation=audio_representation['type'], representation_params=audio_representation)

            n_steps = int(np.ceil(loader.num() / batch_size))

            batch_generator = lambda: batch_load_audio_file_data(loader=loader, batch_size=batch_size, start_idx=last_processed_batch+1, logger=logger)
        else:
            raise ValueError("Unsupported file type. Only '.h5' and '.wav' files are supported.")
    else:
        # is an audio folder
        # recursively getting a list of all files in folder that ends with .wav
        audio_files = [audio_file for audio_file in audio_path.rglob('*.wav')]

        if file_list is not None:
            file_list = pd.read_csv(file_list, header=None)[0].values.tolist()
            # filtering all the files paths to only process the ones we specified in file_list
            audio_files = [row for row in audio_files if row.rsplit(os.sep,1)[1] in file_list] 
    
        # initialize audio loader
        # The audio loader will segment an audio recording into segments of size "duration"
        # This is a generator that yields each segment as a spectrogram
        loader = AudioFrameLoader(duration=audio_representation["duration"], pad=False, step=step_size, batch_size=1,
            filename=audio_files, representation=audio_representation['type'], representation_params=audio_representation)

        n_steps = int(np.ceil(loader.num() / batch_size))

        batch_generator = lambda: batch_load_audio_file_data(loader=loader, batch_size=batch_size, start_idx=last_processed_batch+1, logger=logger)

    logger = KetosLogger('ketos-run', str(Path(output_folder, log_file)), mode=mode, format='yaml')

    # Log the initial parameters
    logger.info("Process started", stdout=True, **{
        'parameters': {
            'model_file': model_file,
            'audio_data': audio_data,
            'file_list': file_list,
            'table_name': table_name,
            'output_folder': str(output_folder),
            'overwrite': overwrite,
            'log_file': log_file,
            'step_size': step_size,
            'batch_size': batch_size,
            'threshold': threshold,
            'merge_detections': merge_detections,
            'buffer': buffer,
            'running_avg': running_avg,
            'highest_score_only': highest_score_only,
            'output_function_arguments': output_function_arguments,
        },
        'n_steps': n_steps,
        'status': 'success'
    })

    filenames = None
    try:
        # the arguments for the chosen generator function have already been passed by the lambda function
        for idx, batch_data in enumerate(batch_generator(), start=last_processed_batch+1):
            successful_process = False
            try:
                filenames = list(dict.fromkeys(str(path) for path in batch_data['filename'])) # getting the unique elements and converting path object to str

                if input_transform_func is not None:
                    data = input_transform_func(batch_data['data'])
                
                batch_predictions = model.run_on_batch(data, transform_input=False, return_raw_output=True) 
                # batch_predictions = model.model.predict_on_batch(data)

                raw_output = {'filename': batch_data['filename'], 'start': batch_data['start'], 'end': batch_data['end'], 'score': batch_predictions}

                with open(output, mode+'b') as f:
                    pickle.dump(raw_output, f)

                # Run the output transform function. The user can customize this function
                output_function(raw_output, output_folder, mode, **output_function_arguments)
                successful_process = True # <-- SET FLAG TO TRUE AFTER SUCCESSFUL PROCESS. We need this flag for when we interrupt the process after successful completion but before the next batch starts
                mode = "a"

                logger.info(f"Step {str(idx)}", **{
                    'step': idx,
                    'files': filenames,
                    'status': 'success'
                })

            except Exception as e:
                if successful_process:
                    idx += 1 # Set to the next index because
                logger.error(f'{e}', stdout=True, **{'status': 'error'})
                raise

    except KeyboardInterrupt as e:
        if filenames is not None:
            if successful_process:
                idx += 1 # Set to the next index because
            logger.error('Process was interrupted by user.', stdout=True, **{
                'step': idx,
                'files': filenames,
                'status': 'interruption'
            })
        raise
    else:
        # This block executes only if the loop completed without encountering a break statement.
        logger.info(f"Process finished. Output saved to: {output_folder}", stdout=True, **{
            'output': {
                'folder': str(output_folder)
            },
            'status': 'success'
        })


def main():
    import argparse
    import warnings
    import ast
    def boolean_string(s):
            if s not in {'False', 'True'}:
                raise ValueError('Not a valid boolean string')
            return s == 'True'

    def tryeval(val):
        # Literal eval does cast type safely. However, doesnt work for str, therefore the try except.
        try:
            val = ast.literal_eval(val)
        except ValueError:
            pass
        return val
        
    class ParseKwargs(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            setattr(namespace, self.dest, dict())
            for value in values:
                key, value = value.split('=')
                getattr(namespace, self.dest)[key] = tryeval(value)

    # parse command-line args
    parser = argparse.ArgumentParser()
    parser.add_argument('model_file', type=str, help='Path to the ketos model file (*.kt)')
    parser.add_argument('audio_data', type=str, help='Path to the audio data to be processed. This can be a path to a single audio file, a directory of audio files, or an HDF5 database.')
    parser.add_argument('--file_list', default=None, type=str, help='A .csv or .txt file where each row (or line) is the name of a file to detect within the audio folder. \
                        By default, all files will be processed. Not relevant if audio_data is an HDF5 file.')
    parser.add_argument('--table_name', default=None, type=str, help="Table name within the HDF5 database where the data is stored. Must start with a foward slash. For instance '/test'. \
                        If not given, the root '/' path will be used. Not relevant if audio_data is a fodler with audio files.")
    parser.add_argument('--output_folder', default=None, type=str, help='Location to output the detections. For instance: detections/')
    parser.add_argument('--overwrite', default=True, type=boolean_string, help='Overwrites the detections, otherwise appends to it.')
    parser.add_argument('--log_file', default="ketos-run.log", type=str, help='Name of the log file to be created/used during the process. Defaults to "ketos-run.log".')
    parser.add_argument('--step_size', default=None, type=float, help='Step size in seconds. If not specified, the step size is set equal to the duration of the audio representation.')
    parser.add_argument('--threshold', default=0.5, type=float, help="The threshold value used to determine the cut-off point for detections. This is a floating-point value between 0 and 1. A detection is considered positive if its score is above this threshold. The default value is 0.5.")
    parser.add_argument('--labels', type=tryeval, default=None, help="List or integer of labels to filter by. Example usage: --labels 1 or --labels [1,2,3]. Defaults to None.")
    parser.add_argument('--merge_detections', default=False, type=boolean_string, 
                    help="A flag indicating whether to merge overlapping detections into a single detection. If set to True, overlapping detections are merged. The default value is False, meaning detections are kept separate.")
    parser.add_argument('--buffer', default=0.0, type=float, 
                    help="The buffer duration to be added to each detection in seconds. This helps to extend the start and end times of each detection to include some context around the detected event. The default value is 0.0, which means no buffer is added.")
    parser.add_argument('--running_avg', default=None, type=int, 
                    help="Compute a running average of the scores over a specified window size in frames. Must be an odd integer.")
    parser.add_argument('--highest_score_only', default=False, type=boolean_string,
                    help='If True, will only return the label associated with the highest score even if more than one passes the threshold. Defaults to False.')
    parser.add_argument('--batch_size', default=32, type=int, help='How many samples will be loaded into memory. Lower this number if you are running into out of memory problems.')
    parser.add_argument('--output_function_arguments', default=None, nargs='*', action=ParseKwargs, help='Output function arguments. If you created a custom output transform function, you can \
        use this option to pass any arguments to it. Usage: --output_function_arguments arg1=value1 arg2=value2')

    args = parser.parse_args()
    try:
        args = parser.parse_args()
        ketos_run(**vars(args))

    except KeyboardInterrupt:
        print("Program was interrupted by the user.")
        sys.exit(EXIT_INTERRUPTION)

    else:
        sys.exit(EXIT_SUCCESS)


if __name__ == "__main__":
    main()


# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Unit tests for the selection_table module within the ketos library
"""
import pytest
import os
import ketos.data_handling.selection_table as st
import pandas as pd
import numpy as np
from io import StringIO
import datetime as dt
import tempfile


current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(os.path.dirname(current_dir),"assets")
path_to_tmp = os.path.join(path_to_assets,'tmp')


def test_trim():
    standard = ['filename','label','start','end','freq_min','freq_max']
    extra = ['A','B','C']
    df = pd.DataFrame(columns=extra)
    df = st.trim(df)
    assert len(df.columns) == 0
    df = pd.DataFrame(columns=standard+extra)
    df = st.trim(df)
    assert sorted(df.columns.values) == sorted(standard)

def test_is_standardized():
    df = pd.DataFrame({'filename':'test.wav','label':[1],'start':[0],'end':[2],'freq_min':[None],'freq_max':[None]})
    df = st.standardize(df)
    assert st.is_standardized(df) == True
    df = pd.DataFrame({'filename':'test.wav','label':[1]})
    df = st.standardize(df)
    assert st.is_standardized(df) == True
    df = pd.DataFrame({'filename':'test.wav','label':[1]})
    assert st.is_standardized(df) == False


def test_empty_annotation_table():
    df = st.empty_annot_table()
    assert len(df) == 0
    assert np.all(df.columns == ['label', 'start', 'end'])
    assert df.index.names == ['filename', 'annot_id']


def test_empty_selection_table():
    df = st.empty_selection_table()
    assert len(df) == 0
    assert np.all(df.columns == ['label', 'start', 'end', 'annot_id'])
    assert df.index.names == ['filename', 'sel_id']


def test_create_label_dict():
    l1 = [0, 'gg', -17, 'whale']
    l2 = [999]
    d = st._create_label_dict(l1, l2)
    ans = {999: -1, 0: 0, 'gg':1, -17: 2, 'whale': 3}
    assert d == ans


def test_create_label_dict_can_handle_nested_list():
    l1 = [[-33, 1, 'boat']]
    l2 = [0, 'gg', [-17, 'whale']]   
    l3 = [999]
    d = st._create_label_dict(l1+l2, l3)
    ans = {-33: 0, 1:0, 'boat': 0, 999: -1, 0: 1, 'gg':2, -17: 3, 'whale': 3}
    assert d == ans


def test_unfold(annot_table_mult_labels):
    res = st.unfold(annot_table_mult_labels)
    ans = pd.DataFrame({'filename':['f0.wav','f0.wav','f1.wav'], 'label':['1','2','3'], 'start':[0,0,1], 'end':[1,1,2]})
    res = res.reset_index(drop=True)[ans.columns]
    pd.testing.assert_frame_equal(ans, res[ans.columns.values])

def test_standardize_with_dataframe_input(basic_dataframe):
    """
    Test standardization with a pandas DataFrame input. Verifies that the function returns
    a DataFrame, it's not empty, and the index includes 'filename' and 'annot_id'.
    """
    result = st.standardize(annotations=basic_dataframe)
    assert isinstance(result, pd.DataFrame)
    assert not result.empty
    assert 'filename' in result.index.names
    assert 'annot_id' in result.index.names

def test_standardize_with_csv_input(temp_csv_file):
    """
    Test standardization with a CSV file input. Verifies that the function returns
    a DataFrame, it's not empty, and cleans up the temporary CSV file after testing.
    """
    result = st.standardize(annotations=temp_csv_file)
    assert isinstance(result, pd.DataFrame)
    assert not result.empty
    os.remove(temp_csv_file)  # Cleanup

def test_standardize_automatic_label_mapping(basic_dataframe):
    """
    Test automatic label mapping during standardization. Verifies that labels are correctly
    mapped to integers, the data columns match expected values, and the 'label_dict' attribute
    contains the correct mappings for 'dolphin' and 'whale'.
    """
    result = st.standardize(annotations=basic_dataframe, labels='auto')
    # Expected data columns, ignoring the index for now
    expected_data = {
        'label': [1, 0],  # Assuming 'dolphin'->'0', 'whale'->'1' after auto mapping
        'start': [0., 5.],
        'end': [10., 15.],
    }
    expected_df = pd.DataFrame(expected_data, index=pd.MultiIndex.from_arrays([
        ['file1', 'file2'],  # filename
        [0, 0]  
    ], names=['filename', 'annot_id']))
  
    # Check data columns match
    pd.testing.assert_frame_equal(result[['label', 'start', 'end']], expected_df[['label', 'start', 'end']])

    # Check the first level of the index (filename) matches
    assert (result.index.get_level_values('filename') == expected_df.index.get_level_values('filename')).all()

    assert sorted(result.attrs['label_dict'].keys()) == ['dolphin', 'whale']

def test_standardize_dict_label_mapping(basic_dataframe):
    """
    Test custom dictionary label mapping during standardization. Verifies that labels are
    correctly mapped according to the provided dictionary.
    """
    custom_labels = {'whale': 1, 'dolphin': 2}
    result = st.standardize(annotations=basic_dataframe, labels=custom_labels)
    assert all(result['label'].map(int).isin([1, 2]))

def test_standardize_no_label_mapping(basic_dataframe):
    """
    Test standardization with no label mapping (labels=None). Verifies that labels
    remain unchanged in the result.
    """
    result = st.standardize(annotations=basic_dataframe, labels=None)
    # Verify labels remain unchanged
    assert sorted(result['label'].unique()) == ['dolphin', 'whale']

def test_standardize_list_of_labels(varied_labels_dataframe):
    """
    Test standardization with a list of labels for mapping. Verifies that labels not
    in the list are mapped to -1, and the 'label_dict' only includes specified labels.
    """
    # Specify only a subset of labels to map
    labels_list = ['whale', 'dolphin']
    result = st.standardize(annotations=varied_labels_dataframe, labels=labels_list)
    # Expect 'shark' to be mapped to -1
    mapped_labels = result['label'].map(int).unique()
    assert sorted(mapped_labels) == [-1, 0, 1]  # 'shark' becomes -1, 'whale' and 'dolphin' are 0 and 1
    # Verify that 'label_dict' only includes specified labels
    assert 'shark' not in result.attrs['label_dict']

def test_standardize_unfold_labels(multi_label_dataframe):
    """
    Test unfolding multiple labels per row during standardization. Verifies that
    rows with multiple labels are correctly unfolded into separate rows.
    """
    result = st.standardize(annotations=multi_label_dataframe, unfold_labels=True, labels=None, label_sep=',')
    assert len(result) == 2  # Two labels should result in two rows

def test_standardize_trim_table(basic_dataframe):
    """
    Test trimming of extra columns during standardization. Verifies that only
    columns relevant to the standardized format are retained in the result.
    """
    basic_dataframe['extra_column'] = 'extra'
    result = st.standardize(annotations=basic_dataframe, trim_table=True)
    assert 'extra_column' not in result.columns

def test_standardize_datetime_format(datetime_dataframe):
    """
    Test handling of datetime format during standardization. Verifies that datetime
    columns are correctly parsed according to the specified format.
    """
    result = st.standardize(annotations=datetime_dataframe, datetime_format='%d_%m_%Y')
    assert pd.to_datetime(result['datetime']).dt.year.eq(2020).all()

def test_standardize_invalid_input_type():
    """
    Test standardization with an invalid input type. Verifies that the function
    raises a ValueError when input is not a DataFrame or a file path.
    """
    with pytest.raises(ValueError):
        st.standardize(annotations=123)

def test_standardize_missing_required_columns():
    """
    Test standardization with missing required columns in the input DataFrame.
    Verifies that the function raises an AssertionError when required columns
    are missing.
    """
    df = pd.DataFrame({'wrong_column': [1, 2]})
    with pytest.raises(AssertionError):
        st.standardize(annotations=df, labels=None)

def test_standardize_datetime():
    """ Test that datetime information is parsed correctly
    """
    filenames = ["empty_HMS_12_ 5_28__DMY_23_ 2_1984.wav",
                 "empty_HMS_12_05_28__DMY_23_03_1924.wav",
                 "subfolder/empty_HMS_23_05_28__DMY_23_ 1_1924.wav"]
    labels = [0,1,2]
    fmt = '*HMS_%H_%M_%S__DMY_%d_%m_%Y*'
    # datetime field not specified (defaults to filename)
    tbl = pd.DataFrame({'filename':filenames,'label':labels})
    res = st.standardize(tbl, datetime_format=fmt)
    assert res['datetime'][0] == dt.datetime(1984, 2, 23, 12, 5, 28) 
    assert res['datetime'][1] == dt.datetime(1924, 3, 23, 12, 5, 28) 
    assert res['datetime'][2] == dt.datetime(1924, 1, 23, 23, 5, 28) 
    # datetime field specified
    t = [os.path.basename(f) for f in filenames]
    tbl = pd.DataFrame({'filename':filenames,'label':labels,'datetime':t})
    res = st.standardize(tbl, datetime_format=fmt)
    assert res['datetime'][0] == dt.datetime(1984, 2, 23, 12, 5, 28) 
    assert res['datetime'][1] == dt.datetime(1924, 3, 23, 12, 5, 28) 
    assert res['datetime'][2] == dt.datetime(1924, 1, 23, 23, 5, 28) 
    tbl = pd.DataFrame({'filename':filenames,'label':labels,'datetime':t})
    res = st.standardize(tbl, datetime_format=fmt)
    assert res['datetime'][0] == dt.datetime(1984, 2, 23, 12, 5, 28) 
    assert res['datetime'][1] == dt.datetime(1924, 3, 23, 12, 5, 28) 
    assert res['datetime'][2] == dt.datetime(1924, 1, 23, 23, 5, 28) 


def test_label_occurrence(annot_table_std):
    df = annot_table_std
    oc = st.label_occurrence(df)
    ans = {-1: 1, 0: 2, 1: 1, 2: 1, 3: 1}
    assert oc == ans


def test_select_center(annot_table_std):
    df = st.standardize(annot_table_std)
    print(df)
    # request length shorter than annotations
    res = st.select(df, length=1, center=True)
    d = '''filename sel_id label  start   end
f0.wav   0           2   1.15  2.15
f0.wav   1           1   4.15  5.15
f1.wav   0           3   2.15  3.15
f1.wav   1           1   5.15  6.15
f2.wav   0           4   3.15  4.15
f2.wav   1           0   6.15  7.15'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(ans, res[ans.columns.values])
    # request length longer than annotations
    res = st.select(df, length=5, center=True)
    d = '''filename sel_id  label  start   end
f0.wav   0           2  -0.85  4.15
f0.wav   1           1   2.15  7.15
f1.wav   0           3   0.15  5.15
f1.wav   1           1   3.15  8.15
f2.wav   0           4   1.15  6.15
f2.wav   1           0   4.15  9.15'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(ans, res[ans.columns.values])


def test_select_ignores_annotations_with_label_minus_1(annot_table_std):
    df = annot_table_std
    df = st.standardize(df)
    res = st.select(df, length=1, center=True)
    assert len(res[res.label==-1]) == 0


def test_select_only_uses_specified_labels(annot_table_std):
    df = st.standardize(annot_table_std)
    res = st.select(df, length=1, center=True, label=[0,2])
    assert np.all(np.isin(res.label.values, [0,2]))


def test_select_avoids_certain_labels(annot_table_std):
    df = st.standardize(annot_table_std)
    res = st.select(df, length=3.3, center=True, label=[1,3], avoid_label=[2])
    assert np.all(np.isin(res.label.values,[1,3]))
    assert len(res) == 2
    res.reset_index(inplace=True)
    res.iloc[0].to_dict() == {'filename': 'f1.wav', 'sel_id': 0, 'label': 3, 'start': 1.0, 'end': 4.3}
    res.iloc[1].to_dict() == {'filename': 'f1.wav', 'sel_id': 0, 'label': 1, 'start': 4.0, 'end': 7.3}
    res = st.select(df, length=3.3, center=True, label=[1,3,0], avoid_label="ALL")
    assert np.all(np.isin(res.label.values,[1,3,0]))
    assert len(res) == 2


def test_select_discards_selections_outside_file(annot_table_std):
    df = st.standardize(annot_table_std)
    files = pd.DataFrame({'filename':['f0.wav','f1.wav','f2.wav'], 'duration':[10,10,10]})
    res = st.select(df, length=7.0, center=True, discard_outside=True, files=files)
    assert len(res) == 3


def test_select_keeps_extra_attrs(annot_table_std):
    annot_table = annot_table_std.copy()
    annot_table['comment'] = ['good', 'tall', 'thin', 'bad', 'hopeless', 'green'] #add extra column
    df = annot_table
    df = st.standardize(df)
    res = st.select(df, length=1, center=True)
    assert 'comment' in res.columns.values


def test_select_enforces_overlap(annot_table_std):
    np.random.seed(3)
    df = annot_table_std
    df = st.standardize(df)
    # requested length: 5.0 sec
    # all annotations have duration: 3.3 sec  (3.3/5.0=0.66)
    length = 5.0
    overlap = 0.5
    df_new = st.select(df, length=length, min_overlap=overlap, keep_id=True)
    t1 = df_new.start.values
    t2 = df_new.end.values
    idx = zip(df_new.index.get_level_values(0), df_new.annot_id)
    df = df.loc[idx]
    t2_orig = df.end.values
    t1_orig = df.start.values
    assert np.all(t2 >= t1_orig + overlap * length)
    assert np.all(t1 <= t2_orig - overlap * length)


def test_select_step(annot_table_std):
    df = annot_table_std
    df = st.standardize(df)
    N = len(df[df['label']!=-1])
    K = len(df[df['label']==5])
    df_new = st.select(df, length=1, center=True, min_overlap=0, step=0.5, keep_id=True)
    M = len(df_new)
    assert M == (N - K) * (2 * int((3.3/2+0.5)/0.5) + 1) + K * (2 * int((3.3/2-0.5)/0.5) + 1)
    df_new = st.select(df, length=1, center=True, min_overlap=0.4, step=0.5)
    M = len(df_new)
    assert M == (N - K) * (2 * int((3.3/2+0.5-0.4)/0.5) + 1) + K * (2 * int((3.3/2-0.5)/0.5) + 1)

def test_select_warning_annotation_error(annot_table_std):
    """ Test that a warning is given when an annotation contains a start time greater than end time
    and that select ignores the annotation
    """
    np.random.seed(2)
    df = annot_table_std
    df["end"][1] = 0.23
    df = st.standardize(df)

    with pytest.warns(UserWarning):
        res = st.select(annotations=df, length=1.0, center=False)

    d ='''filename sel_id label start end                            
f0.wav   0           2  1.002788  2.002788
f0.wav   1           1  3.059630  4.059630
f1.wav   1           1  5.264224  6.264224
f2.wav   0           4  3.001242  4.001242
f2.wav   1           0  5.966846  6.966846'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(ans, res[ans.columns.values])


def test_time_shift():
    row = pd.Series({'label':3.00,'start':0.00,'end':3.30,'annot_id':0.00,'length':3.30,'start_new':-0.35})
    res = st.time_shift(annot=row, time_ref=row['start_new'], length=4.0, min_overlap=0.8, step=0.5)
    d = '''label  start  end  annot_id  length  start_new
0    3.0    0.0  3.3       0.0     3.3      -1.35
1    3.0    0.0  3.3       0.0     3.3      -0.85
2    3.0    0.0  3.3       0.0     3.3      -0.35
3    3.0    0.0  3.3       0.0     3.3       0.15
4    3.0    0.0  3.3       0.0     3.3       0.65'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0])
    pd.testing.assert_frame_equal(ans, res[ans.columns.values])

def test_time_shift_negative_length_raises_exception():
    """ Test that the assertion error for a length smaller than 0 is thrown
    """
    annot = {'filename':'file1.wav', 'label':1, 'start':12.0, 'end':14.0}

    with pytest.raises(AssertionError):
        st.time_shift(annot, time_ref=13.0, length=-1.0, step=0.2, min_overlap=0.5)

def test_time_shift_positive_length_doesnt_raises_exception():
    """ Test that the assertion error is not thrown for good values
    """
    annot = {'filename':'file1.wav', 'label':1, 'start':12.0, 'end':14.0}

    try:
        st.time_shift(annot, time_ref=13.0, length=1.0, step=0.2, min_overlap=0.5)
    except AssertionError as exc:
        assert False, f"'time_shift' raised an exception {exc}"

def test_select_with_varying_overlap(annot_table_std):
    """ Test that the number of selections increases as the 
        minimum required overlap is reduced"""
    df = st.standardize(annot_table_std)
    # request length shorter than annotations
    num_sel = []
    for min_overlap in np.linspace(1.0, 0.0, 11):
        res = st.select(df, length=1.0, step=0.5, center=True, min_overlap=min_overlap)
        num_sel.append(len(res))
    
    assert np.all(np.diff(num_sel) >= 0)
    # request length longer than annotations
    num_sel = []
    for min_overlap in np.linspace(1.0, 0.0, 11):
        res = st.select(df, length=4.0, step=0.5, center=True, min_overlap=min_overlap)
        num_sel.append(len(res))

    assert np.all(np.diff(num_sel) >= 0)


def test_create_rndm_selections_with_label(file_duration_table):
    np.random.seed(1)
    dur = file_duration_table 
    num = 5
    df_bgr = st.create_rndm_selections(files=dur, length=2.0, num=num, label=6)
    assert len(df_bgr) == num
    # assert selections have uniform length
    assert np.all(np.isclose(df_bgr.end.values - df_bgr.start.values, 2.0, atol=1e-9))
    # assert all selection have label = 6
    assert np.all(df_bgr.label.values == 6)




def test_create_rndm_selections_with_empty_annot(file_duration_table):
    """ Test if can generate a random selection with an empty annotation table"""
    np.random.seed(1)
    dur = file_duration_table 
    num = 5
    df = st.empty_annot_table()
    df_bgr = st.create_rndm_selections(annotations=df, files=dur, length=2.0, num=num)
    assert len(df_bgr) == num
    # assert selections have uniform length
    assert np.all(np.isclose(df_bgr.end.values - df_bgr.start.values, 2.0, atol=1e-9))
    # assert all selection have label = 0
    assert np.all(df_bgr.label.values == 0)


def test_create_rndm_selections_with_empty_file_duration_table():
    """ Test if can generate a random selection with an empty annotation table"""
    np.random.seed(1)
    num = 5
    df = st.empty_annot_table()
    dur = pd.DataFrame(columns=['filename','duration'])
    df_bgr = st.create_rndm_selections(annotations=df, files=dur, length=2.0, num=num)
    assert len(df_bgr) == 0


def test_create_rndm_selections_with_annot(annot_table_std, file_duration_table):
    """ Test if can generate a random selection while avoiding annotated regions of the recording"""
    np.random.seed(1)
    df = st.standardize(annot_table_std)
    dur = file_duration_table 
    num = 5
    df_bgr = st.create_rndm_selections(annotations=df, files=dur, label=10, length=2.0, num=num)
    assert len(df_bgr) == num
    # assert selections have uniform length
    assert np.all(np.abs(df_bgr.end.values - df_bgr.start.values - 2.0) < 1e-9)
    # assert all selection have label = 10
    assert np.all(df_bgr.label.values == 10)
    # assert selections do not overlap with any annotations
    for bgr_idx, bgr_sel in df_bgr.iterrows():
        start_bgr = bgr_sel.start
        end_bgr = bgr_sel.end
        fname = bgr_idx[0]
        q = st.query(df, start=start_bgr, end=end_bgr, filename=fname)
        assert len(q) == 0


def test_create_rndm_keeps_misc_cols(annot_table_std, file_duration_table):
    """ Check that the random selection creation method keeps 
        any miscellaneous columns"""
    np.random.seed(1)
    df = st.standardize(annot_table_std)
    dur = file_duration_table 
    dur['extra'] = 'testing'
    df_bgr = st.create_rndm_selections(annotations=df, files=dur, length=2.0, num=5)
    assert np.all(df_bgr['extra'].values == 'testing')
    df_bgr = st.create_rndm_selections(annotations=df, files=dur, length=2.0, num=5, trim_table=True)
    assert 'extra' not in df_bgr.columns.values.tolist()


def test_create_rndm_files_missing_duration(annot_table_std, file_duration_table):
    """ Check that the random selection creation method works even when 
        some of the files are missing from the file duration list"""
    np.random.seed(1)
    df = st.standardize(annot_table_std)
    dur = file_duration_table.drop(0) 
    df_bgr = st.create_rndm_selections(annotations=df, files=dur, length=2.0, num=11)


def test_create_rndm_nonzero_offset(annot_table_std, file_duration_table):
    """ Check that the random selection creation method works when file duration 
        table includes offsets"""
    np.random.seed(1)
    df = st.standardize(annot_table_std)
    file_duration_table['offset'] = [1, 2, 3, 4, 5, 6]
    df_bgr = st.create_rndm_selections(annotations=df, files=file_duration_table, length=2.0, num=11)


def test_create_rndm_selections_no_overlap(annot_table_std, file_duration_table):
    """ Check that random selections have no overlap"""
    np.random.seed(1)
    df = st.standardize(annot_table_std)
    dur = file_duration_table 
    num = 30
    df_bgr = st.create_rndm_selections(annotations=df, files=dur, length=2.0, num=num)
    num_overlap = 0
    for idx,row in df_bgr.iterrows():
        q = st.query(df_bgr, filename=idx[0], start=row['start'], end=row['end'])
        num_overlap += len(q) - 1
    
    assert num_overlap > 0

    df_bgr = st.create_rndm_selections(annotations=df, files=dur, length=2.0, num=num, no_overlap=True)
    num_overlap = 0
    for idx,row in df_bgr.iterrows():
        q = st.query(df_bgr, filename=idx[0], start=row['start'], end=row['end'])
        num_overlap += len(q) - 1
    
    assert num_overlap == 0


def test_select_by_segmenting(annot_table_std, file_duration_table):
    a = st.standardize(annot_table_std)
    f = file_duration_table
    sel = st.select_by_segmenting(f, length=5.1, annotations=a, step=4.0, discard_empty=True, pad=True)
    # check selection table
    d = '''filename sel_id start  end
f0.wav   0         0.0  5.1
f0.wav   1         4.0  9.1
f1.wav   0         0.0  5.1
f1.wav   1         4.0  9.1
f2.wav   0         0.0  5.1
f2.wav   1         4.0  9.1
f2.wav   2         8.0 13.1'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(ans, sel[0][ans.columns.values])
    # check annotation table
    d = '''filename sel_id annot_id label  start  end
f0.wav   0      0             2         0.0        3.3
f0.wav   0      1             1         3.0        6.3
f0.wav   1      1             1        -1.0        2.3
f1.wav   0      0             3         1.0        4.3
f1.wav   0      1             1         4.0        7.3
f1.wav   1      0             3        -3.0        0.3
f1.wav   1      1             1         0.0        3.3
f2.wav   0      0             4         2.0        5.3
f2.wav   0      1             0         5.0        8.3
f2.wav   1      0             4        -2.0        1.3
f2.wav   1      1             0         1.0        4.3
f2.wav   2      1             0        -3.0        0.3'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1,2])
    pd.testing.assert_frame_equal(ans, sel[1][ans.columns.values])


def test_select_by_segmenting_keep_only_empty(annot_table_std, file_duration_table):
    a = st.standardize(annot_table_std)
    f = file_duration_table
    sel = st.select_by_segmenting(f, length=16.0, annotations=a, step=4.0, keep_only_empty=True, pad=False)
    # check selection table
    d = '''filename sel_id start  end
f0.wav   2         8.0  24.0
f0.wav   3        12.0  28.0
f1.wav   2         8.0  24.0
f1.wav   3        12.0  28.0
f2.wav   3        12.0  28.0
f2.wav   4        16.0  32.0
f3.wav   0         0.0  16.0
f3.wav   1         4.0  20.0
f3.wav   2         8.0  24.0
f3.wav   3        12.0  28.0
f3.wav   4        16.0  32.0
f4.wav   0         0.0  16.0
f4.wav   1         4.0  20.0
f4.wav   2         8.0  24.0
f4.wav   3        12.0  28.0
f4.wav   4        16.0  32.0
f5.wav   0         0.0  16.0
f5.wav   1         4.0  20.0
f5.wav   2         8.0  24.0
f5.wav   3        12.0  28.0
f5.wav   4        16.0  32.0'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(ans, sel[ans.columns.values])


def test_select_by_segmenting_label_empty(annot_table_std, file_duration_table):
    a = st.standardize(annot_table_std)
    f = file_duration_table
    sel = st.select_by_segmenting(f, length=16.0, annotations=a, step=4.0, keep_only_empty=True, label_empty=6, pad=False)
    # check selection table
    d = '''filename sel_id start end label
f0.wav   2         8.0  24.0  6
f0.wav   3        12.0  28.0  6
f1.wav   2         8.0  24.0  6
f1.wav   3        12.0  28.0  6
f2.wav   3        12.0  28.0  6
f2.wav   4        16.0  32.0  6
f3.wav   0         0.0  16.0  6
f3.wav   1         4.0  20.0  6
f3.wav   2         8.0  24.0  6
f3.wav   3        12.0  28.0  6
f3.wav   4        16.0  32.0  6
f4.wav   0         0.0  16.0  6
f4.wav   1         4.0  20.0  6
f4.wav   2         8.0  24.0  6
f4.wav   3        12.0  28.0  6
f4.wav   4        16.0  32.0  6
f5.wav   0         0.0  16.0  6
f5.wav   1         4.0  20.0  6
f5.wav   2         8.0  24.0  6
f5.wav   3        12.0  28.0  6
f5.wav   4        16.0  32.0  6'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(ans, sel[ans.columns.values])


def test_select_by_segmenting_avoid_label(annot_table_std, file_duration_table):
    a = st.standardize(annot_table_std)
    f = file_duration_table[:2]
    sel, _ = st.select_by_segmenting(f, length=16.0, annotations=a, step=4.0, avoid_label=[2], pad=False)
    # check selection table
    d = '''filename sel_id start end
f0.wav   1         4.0  20.0
f0.wav   2         8.0  24.0
f0.wav   3        12.0  28.0
f1.wav   0         0.0  16.0
f1.wav   1         4.0  20.0
f1.wav   2         8.0  24.0
f1.wav   3        12.0  28.0'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(ans, sel[ans.columns.values])


def test_query_labeled(annot_table_std):
    df = st.standardize(annot_table_std)
    df = st.select(df, length=1, center=True)
    # query for file that does not exist
    q = st.query_labeled(df, filename='fff.wav')
    assert len(q) == 0
    # query for 1 file
    q = st.query_labeled(df, filename='f1.wav')
    d = '''filename sel_id label  start   end                   
f1.wav  0           3   2.15  3.15
f1.wav  1           1   5.15  6.15'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(q, ans[q.columns.values])
    # query for 1 file, and 1 that does not exist
    q = st.query_labeled(df, filename=['f1.wav','fff.wav'])
    d = '''filename sel_id label  start   end                   
f1.wav   0           3   2.15  3.15
f1.wav   1           1   5.15  6.15'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(q, ans[q.columns.values])
    # query for 2 files
    q = st.query_labeled(df, filename=['f1.wav','f2.wav'])
    d = '''filename sel_id label  start   end                            
f1.wav   0           3   2.15  3.15
f1.wav   1           1   5.15  6.15
f2.wav   0           4   3.15  4.15
f2.wav   1           0   6.15  7.15'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(q, ans[q.columns.values])
    # query for labels
    q = st.query_labeled(df, label=[1,4])
    d = '''filename sel_id label  start   end                   
f0.wav   1           1   4.15  5.15
f1.wav   1           1   5.15  6.15
f2.wav   0           4   3.15  4.15'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(q, ans[q.columns.values])
    # query for label that does not exist
    q = st.query_labeled(df, label=99)
    assert len(q) == 0


def test_query_annotated(annot_table_std, file_duration_table):
    a = st.standardize(annot_table_std)
    f = file_duration_table
    sel = st.select_by_segmenting(f, length=5.1, annotations=a, step=4.0, discard_empty=True, pad=True)
    # query for 1 file
    q1, q2 = st.query_annotated(sel[0], sel[1], label=[1,3])
    d = '''filename sel_id start  end
f0.wav   0         0.0  5.1
f0.wav   1         4.0  9.1
f1.wav   0         0.0  5.1
f1.wav   1         4.0  9.1'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(q1, ans[q1.columns.values])
    d = '''filename sel_id annot_id label  start  end  
f0.wav   0      1             1    3.0  6.3
f0.wav   1      1             1   -1.0  2.3
f1.wav   0      0             3    1.0  4.3
f1.wav   0      1             1    4.0  7.3
f1.wav   1      0             3   -3.0  0.3
f1.wav   1      1             1    0.0  3.3'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1,2])
    pd.testing.assert_frame_equal(q2, ans[q2.columns.values])


def test_file_duration_table(five_time_stamped_wave_files):
    """ Test that we can generate a file duration table""" 
    df = st.file_duration_table(five_time_stamped_wave_files)
    d = '''filename,duration
empty_HMS_12_ 5_ 0__DMY_23_ 2_84.wav,0.5
empty_HMS_12_ 5_ 1__DMY_23_ 2_84.wav,0.5
empty_HMS_12_ 5_ 2__DMY_23_ 2_84.wav,0.5
empty_HMS_12_ 5_ 3__DMY_23_ 2_84.wav,0.5
empty_HMS_12_ 5_ 4__DMY_23_ 2_84.wav,0.5'''
    ans = pd.read_csv(StringIO(d))
    pd.testing.assert_frame_equal(df, ans[df.columns.values])

def test_file_duration_table_datetime(five_time_stamped_wave_files):
    """ Test that we can generate a file duration table with datetime column""" 
    df = st.file_duration_table(five_time_stamped_wave_files, datetime_format="*HMS_%H_%M_%S__DMY_%d_%m_%y.wav")
    d = '''filename,duration,datetime
empty_HMS_12_ 5_ 0__DMY_23_ 2_84.wav,0.5,1984-02-23 12:05:00
empty_HMS_12_ 5_ 1__DMY_23_ 2_84.wav,0.5,1984-02-23 12:05:01
empty_HMS_12_ 5_ 2__DMY_23_ 2_84.wav,0.5,1984-02-23 12:05:02
empty_HMS_12_ 5_ 3__DMY_23_ 2_84.wav,0.5,1984-02-23 12:05:03
empty_HMS_12_ 5_ 4__DMY_23_ 2_84.wav,0.5,1984-02-23 12:05:04'''
    ans = pd.read_csv(StringIO(d)).astype({'datetime':np.datetime64})
    pd.testing.assert_frame_equal(df, ans[df.columns.values])


def test_random_choice(annot_table_std, file_duration_table):
    a = st.standardize(annot_table_std)
    f = file_duration_table
    sel = st.select_by_segmenting(f, length=16.0, annotations=a, step=4.0, keep_only_empty=True, pad=False)
    sel = st.random_choice(sel, siz=3)
    assert len(sel) == 3
    a = st.random_choice(a, siz=2)
    assert len(a) == 2


def test_aggregate_duration(annot_table_std):
    df = st.standardize(annot_table_std)
    agg_dur = st.aggregate_duration(df)
    assert agg_dur == 18.9
    agg_dur = st.aggregate_duration(df, label=1)
    assert agg_dur == 6.6


import pytest
import numpy as np
import tensorflow as tf
from ketos.neural_networks.dev_utils.nn_interface import RecipeCompat
from ketos.neural_networks.inception import ConvBatchNormRelu, InceptionBlock, InceptionArch, InceptionInterface
#from ketos.neural_networks.dev_utils.losses import FScoreLoss
from ketos.data_handling.data_feeding import BatchGenerator
#from ketos.neural_networks.dev_utils.metrics import Precision, Recall, Accuracy, FScore
import os
import tables
import json

current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(os.path.dirname(current_dir),"assets")
path_to_tmp = os.path.join(path_to_assets,'tmp')

@pytest.fixture
def recipe_dict():
    recipe = {'interface': 'InceptionInterface',
                'n_blocks':3,
                'n_classes':2,
                'initial_filters':16,
                'optimizer': {'recipe_name':'Adam', 'parameters': {'learning_rate':0.005}},
                'loss_function': {'recipe_name':'CategoricalCrossentropy', 'parameters':{}},  
                'metrics': [{'recipe_name':'CategoricalAccuracy', 'parameters':{}}]
    
    }

    return recipe

@pytest.fixture
def recipe():
    recipe = {'interface': 'InceptionInterface',
                'n_blocks':3,
                'n_classes':2,
                'initial_filters':16,
                'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                'loss_function': RecipeCompat('CategoricalCrossentropy', tf.keras.losses.CategoricalCrossentropy),  
                'metrics': [RecipeCompat('CategoricalAccuracy',tf.keras.metrics.CategoricalAccuracy),
                            ],
                }

    return recipe

def test_ConvBatchNormRelu():
    layer = ConvBatchNormRelu(n_filters=16, filter_shape=3, strides=1, padding='same')

    
    assert len(layer.layers[0].layers) == 3
    assert isinstance(layer.layers[0].layers[0], tf.keras.layers.Conv2D)
    assert isinstance(layer.layers[0].layers[1], tf.keras.layers.BatchNormalization)
    assert isinstance(layer.layers[0].layers[2], tf.keras.layers.ReLU)


def test_InceptionBlock():
    block = InceptionBlock(n_filters=16, strides=1)

    assert len(block.layers) == 6
    assert isinstance(block.layers[0], ConvBatchNormRelu)
    assert isinstance(block.layers[1], ConvBatchNormRelu)
    assert isinstance(block.layers[2], ConvBatchNormRelu)
    assert isinstance(block.layers[3], ConvBatchNormRelu)
    assert isinstance(block.layers[4], tf.keras.layers.MaxPooling2D)
    assert isinstance(block.layers[5], ConvBatchNormRelu)


def test_InceptionArch():
    inception = InceptionArch(n_blocks=3, n_classes=2, initial_filters=16)

    assert len(inception.layers) == 5
    assert isinstance(inception.layers[0], ConvBatchNormRelu)
    assert len(inception.layers[1].layers) == 6
    for block in inception.layers[1].layers:
        assert isinstance(block, InceptionBlock)
    assert isinstance(inception.layers[2], tf.keras.layers.GlobalAveragePooling2D)
    assert isinstance(inception.layers[3], tf.keras.layers.Dense)
    assert isinstance(inception.layers[4], tf.keras.layers.Softmax)


def test_InceptionInterface_build_from_recipe(recipe):
    inception = InceptionInterface._build_from_recipe(recipe)

    assert inception.optimizer.recipe_name == recipe['optimizer'].recipe_name
    assert inception.optimizer.instance.__class__ == recipe['optimizer'].instance.__class__
    assert inception.optimizer.args == recipe['optimizer'].args

    assert inception.loss_function.recipe_name == recipe['loss_function'].recipe_name
    assert inception.loss_function.instance.__class__ == recipe['loss_function'].instance.__class__
    assert inception.loss_function.args == recipe['loss_function'].args

    assert inception.metrics[0].recipe_name == recipe['metrics'][0].recipe_name
    assert inception.metrics[0].instance.__class__ == recipe['metrics'][0].instance.__class__
    assert inception.metrics[0].args == recipe['metrics'][0].args

    assert inception.n_blocks == recipe['n_blocks']
    assert inception.initial_filters == recipe['initial_filters']
    assert inception.n_classes ==  recipe['n_classes']    

def test_extract_recipe_dict(recipe, recipe_dict):
    inception = InceptionInterface._build_from_recipe(recipe)
    written_recipe = inception._extract_recipe_dict()

    assert written_recipe == recipe_dict

def test_read_recipe_file(recipe, recipe_dict):
    path_to_recipe_file = os.path.join(path_to_tmp, "test_inception_recipe.json")
    inception = InceptionInterface._build_from_recipe(recipe)
    written_recipe = inception._extract_recipe_dict()
    inception.save_recipe_file(path_to_recipe_file)

    #Read recipe as a recipe dict
    read_recipe = inception._read_recipe_file(path_to_recipe_file,return_recipe_compat=False)
    assert read_recipe == recipe_dict

    #Read recipe as a recipe dict with RecipeCompat objects
    read_recipe = inception._read_recipe_file(path_to_recipe_file,return_recipe_compat=True)
    assert read_recipe['optimizer'].recipe_name ==recipe['optimizer'].recipe_name
    assert read_recipe['optimizer'].instance.__class__ == recipe['optimizer'].instance.__class__
    assert read_recipe['optimizer'].args == recipe['optimizer'].args

    assert read_recipe['loss_function'].recipe_name == recipe['loss_function'].recipe_name
    assert read_recipe['loss_function'].instance.__class__ == recipe['loss_function'].instance.__class__
    assert read_recipe['loss_function'].args == recipe['loss_function'].args
    
    assert read_recipe['metrics'][0].recipe_name == recipe['metrics'][0].recipe_name
    assert read_recipe['metrics'][0].instance.__class__ == recipe['metrics'][0].instance.__class__
    assert read_recipe['metrics'][0].args == recipe['metrics'][0].args

    assert read_recipe['n_blocks'] == recipe['n_blocks']
    assert read_recipe['initial_filters'] == recipe['initial_filters']
    assert read_recipe['n_classes'] ==  recipe['n_classes']  
    

def test_train_Inception(sample_data):
    data, labels = sample_data
    inception = InceptionInterface() #default inception network
    train_generator = BatchGenerator(batch_size=5, x=data, y=labels, shuffle=True)
    val_generator = BatchGenerator(batch_size=5, x=data, y=labels, shuffle=True)

    inception.train_generator = train_generator
    inception.val_generator = val_generator

    inception.train_loop(2, checkpoint_freq=0)



    



